import sys
from Crypto.Cipher import AES
from Crypto.Cipher._mode_cbc import CbcMode
from Crypto.Cipher._mode_ccm import CcmMode
from Crypto.Cipher._mode_ecb import EcbMode
from Crypto.Cipher._mode_cfb import CfbMode
from Crypto.Cipher._mode_ctr import CtrMode
from Crypto.Cipher._mode_eax import EaxMode
from Crypto.Cipher._mode_ofb import OfbMode
from Crypto.Cipher._mode_openpgp import OpenPgpMode
from Crypto.Cipher._mode_siv import SivMode
from Crypto.Random import get_random_bytes
from enum import StrEnum
from pathlib import Path
from pypomes_core import (
    APP_PREFIX,
    file_get_data, exc_format, env_get_enum
)
from typing import Literal, Final

AesMode = type[CbcMode | CcmMode | CfbMode, CtrMode | EaxMode | EcbMode | OfbMode | OpenPgpMode | SivMode]


class SymmetricMode(StrEnum):
    """
    Chaining modes for AES.
    """
    CCM = "CCM"          # Counter with CBC-MAC
    EAX = "EAX"          # Authenticated Encryption with Associated Data
    GCM = "GCM"          # Galois/Counter
    OCB = "OCB"          # Offset CodeBook
    SIV = "SIV"          # Synthetic Initialization Vector


CRYPTO_DEFAULT_SYMMETRIC_MODE: Final[SymmetricMode] = \
    env_get_enum(key=f"{APP_PREFIX}_CRYPTO_DEFAULT_SYMMETRIC_MODE",
                 enum_class=SymmetricMode,
                 def_value=SymmetricMode.EAX)


def crypto_aes_encrypt(plaintext: Path | str | bytes,
                       key: bytes,
                       nonce: bytes = None,
                       header: bytes = None,
                       mode: SymmetricMode = SymmetricMode.EAX,
                       errors: list[str] = None) -> tuple[bytes, bytes, bytes]:
    """
    Symmetrically encrypt *plaintext* using the given *key*, and the chaining mode specified in *mode*.

    The *AES* (Advanced Encryption Standard) symmetric block cipher, as standardized by *NIST*
    (National Institute of Standards and Technology), is used. *AES* is the *de facto* standard
    for symmetric encryption.

    The nature of *plaintext* depends on its data type:
      - type *bytes*: *plaintext* holds the data (used as is)
      - type *str*: *plaintext* holds the data (used as utf8-encoded)
      - type *Path*: *plaintext* is a path to a file holding the data

    Besides providing guarantees over the *confidentiality* of the message, the cipher also provides guarantees over
    its *integrity* (i.e., it allows the receiver to establish whether the *ciphertext* was modified in transit).
    This is accomplished by combining *AEAD* (Authenticated Encryption with Associated Data) with one of
    the following supported chaining modes:
      - *CCM* (Counter with CBC-MAC)
      - *EAX* (Authenticated Encryption with Associated Data)
      - *GCM* (Galois/Counter)
      - *SIV* (Synthetic Initialization Vector)
      - *OCB* (Offset CodeBook)

    The mandatory *key* must be 16 (*AES-128*), 24 (*AES-192*), or 32 (*AES-256*) bytes long.
    For chaining mode *SIV* only, it doubles to 32, 48, or 64 bytes.

    A cryptographic *number once* value (*nonce*) may optionally be provided, but it should never be reused
    for any other encryption done with *key*. Requirements on its byte length depends upon the chaining mode used:
      - *CCM*: range [7 - 13] (11 recommended)
      - *EAX*: no restrictions (16 recommended)
      - *GCM*: no restrictions (16 recommended)
      - *SIV*: no restrictions (16 recommended)
      - *OCB*: range [1 - 15] (15 recommended)
    If *nonce* is not provided, the encryption is rendered deterministic for chaining mode *SIV*.
    For the other modes, a random byte string of the recommended length is automatically created.

    On successful encryption, a tuple with three byte strings is returned:
      - the encrypted message (*ciphertext*)
      - the *number once* value used (*nonce*)
      - the final authentication tag created (*MAC* tag)
    On decryption, the *ciphertext*, *key*, *header*, *nonce* and the *MAC* tag, as well as the chaining
    mode used, must be provided. The *MAC* tag allows the cipher to provide cryptographic *authentication*.

    :param plaintext: the message to encrypt
    :param key: the cryptographic key
    :param nonce: the optional cryptographic *number once* value
    :param header: the optional message header
    :param mode: the chaining mode to use (defaults to *EAX*)
    :param errors: incidental error messages (may be non-empty)
    :return: a tuple containing the encrypted message, and the *nonce* and *MAC* tag used, or *None* on error
    """
    # initialize the return variable
    result: tuple[bytes, bytes, bytes] | None = None

    # obtain the data for encryption
    plaindata: bytes = file_get_data(file_data=plaintext)

    # build the cypher
    # noinspection PyTypeChecker
    cipher: AesMode = AES.new(key=key,
                              mode=__to_symmetric_mode(mode))
    if header:
        cipher.update(header)
    cipher.nonce = nonce or get_random_bytes(16)

    # encrypt the data
    try:
        ciphertext, tag = cipher.encrypt_and_digest(plaintext=plaindata)
        result = (ciphertext, cipher.nonce, tag)
    except Exception as e:
        if isinstance(errors, list):
            exc_err: str = exc_format(exc=e,
                                      exc_info=sys.exc_info())
            errors.append(exc_err)

    return result


def crypto_aes_decrypt(ciphertext: Path | str | bytes,
                       key: bytes,
                       mac_tag: bytes,
                       nonce: bytes,
                       header: bytes = None,
                       mode: SymmetricMode = SymmetricMode.EAX,
                       errors: list[str] = None) -> bytes:
    """
    Symmetrically decrypt *ciphertext* using the given *key*, and the mode of operation specified in *mode*.

    It is assumed that the *AES* (Advanced Encryption Standard) symmetric block cipher, as standardized by *NIST*
    (National Institute of Standards and Technology), was used to generate *ciphertext*.
    *AES* is the *de facto* standard for symmetric encryption.

    The *key*, *header*, and *mode* must be the same ones used to generate *ciphertext*. The *number once*
    (*nonce*) and the authentication *MAC* tag (*mac_tag*) must be the ones returned by the encryption process.
    The *mac_tag* allows the cipher to provide cryptographic *authentication*.

    The nature of *ciphertext* depends on its data type:
      - type *bytes*: *ciphertext* holds the data (used as is)
      - type *str*: *ciphertext* holds the data (used as utf8-encoded)
      - type *Path*: *ciphertext* is a path to a file holding the data

    Besides providing guarantees over the *confidentiality* of the message, the cipher also provides guarantees over
    its *integrity* (i.e., it allows the receiver to establish whether the *ciphertext* was modified in transit).
    This is accomplished by combining *AEAD* (Authenticated Encryption with Associated Data) with one of
    the following supported chaining modes:
      - *CCM* (Counter with CBC-MAC)
      - *EAX* (Authenticated Encryption with Associated Data)
      - *GCM* (Galois/Counter)
      - *SIV* (Synthetic Initialization Vector)
      - *OCB* (Offset CodeBook)

    :param ciphertext: the message to decrypt
    :param key: the cryptographic key
    :param mac_tag: the *MAC* authentication tag
    :param nonce: the cryptographic *number once* value
    :param header: the optional message header
    :param mode: the chaining mode to use (defaults to *EAX*)
    :param errors: incidental error messages (may be non-empty)
    :return: the decrypted message, or *None* on error
    """
    # initialize the return variable
    result: bytes | None = None

    # obtain the data for decryption
    cipherdata: bytes = file_get_data(file_data=ciphertext)

    # build the cypher
    # noinspection PyTypeChecker
    cipher: AesMode = AES.new(key=key,
                              mode=__to_symmetric_mode(mode))
    if header:
        cipher.update(header)
    cipher.nonce = nonce

    # decrypt the data
    try:
        result = cipher.decrypt_and_verify(ciphertext=cipherdata,
                                           received_mac_tag=mac_tag)
    except Exception as e:
        if isinstance(errors, list):
            exc_error: str = exc_format(exc=e,
                                        exc_info=sys.exc_info())
            errors.append(exc_error)

    return result


def __to_symmetric_mode(tag: SymmetricMode) -> Literal[8, 9, 10, 11, 12]:
    """
    Convert the given SymmetricMode *tag* to its internal literal value.

    :param tag: the SymmetricMode value to convert
    :return: the corresponding internal literal value
    """
    result: Literal[8, 9, 10, 11, 12]
    match tag:
        case SymmetricMode.CCM:
            result = AES.MODE_CCM
        case SymmetricMode.SIV:
            result = AES.MODE_SIV
        case SymmetricMode.GCM:
            result = AES.MODE_GCM
        case SymmetricMode.OCB:
            result = AES.MODE_OCB
        case _:  # SymmetricMode.EAX:
            result = AES.MODE_EAX

    return result



import sys
import certifi
import requests
import struct
from contextlib import suppress
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
from cryptography.hazmat.primitives.serialization import (
    BestAvailableEncryption, NoEncryption, KeySerializationEncryption, Encoding, PrivateFormat
)
from cryptography.hazmat.primitives.serialization.pkcs12 import serialize_key_and_certificates
from cryptography.x509.ocsp import OCSPRequestBuilder, load_der_ocsp_response
from cryptography.x509.oid import NameOID
from datetime import UTC, datetime, timedelta
from logging import Logger
from pathlib import Path
from pypomes_core import exc_format
from typing import Any, Literal


def cert_make_certificate(common_name: str,
                          organization: str,
                          locality: str,
                          province: str,
                          country: str,
                          valid_from: datetime = None,
                          valid_for: int = 365,
                          fmt: Literal["der", "pem", "x509"] = "der") -> (tuple[str, str] |
                                                                          tuple[bytes, bytes] |
                                                                          tuple[x509.Certificate, RSAPrivateKey]):
    """
    Generate a self-signed x509 digital certificate/private key set.

    The return format of the certificate and private key is specified in *fmt*:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*
        - *x509*: *cryptography*'s *Certificate* and *RSAPrivateKey* objects

    :param common_name: the certificates's common name, typically the holder's identification
    :param organization: the reference organization
    :param locality: the reference locality, typically the name of a city
    :param province: the reference province, typically the name of a state or province
    :param country: the two-letter *Country Code Top-Level Domain* (ccTLD)
    :param valid_from: the certificate's validation start (defaults to now)
    :param valid_for: the certificate's validation period, in days (defaults to one year)
    :param fmt: the output format to use for the certificate and private key (defaults to *der*)
    :return: a self-signed x509 digital certificate, along with the corresponding private key
    """
    # generate RSA private key
    private_key: RSAPrivateKey = rsa.generate_private_key(public_exponent=65537,
                                                          key_size=2048)
    # build 'subject' and 'ìssuer'
    # (they are the same for self-signed certificates)
    subject_issuer: x509.Name = x509.Name([x509.NameAttribute(oid=NameOID.COUNTRY_NAME,
                                                              value=country.upper()),
                                           x509.NameAttribute(oid=NameOID.STATE_OR_PROVINCE_NAME,
                                                              value=province),
                                           x509.NameAttribute(oid=NameOID.LOCALITY_NAME,
                                                              value=locality),
                                           x509.NameAttribute(oid=NameOID.ORGANIZATION_NAME,
                                                              value=organization),
                                           x509.NameAttribute(oid=NameOID.COMMON_NAME,
                                                              value=common_name)])
    # build the validity period
    if valid_from is None:
        valid_from = datetime.now(tz=UTC)
    valid_to: datetime = valid_from + timedelta(days=valid_for)

    # build the certificate
    certificate: x509.Certificate = (x509.CertificateBuilder()
                                         .subject_name(name=subject_issuer)
                                         .issuer_name(name=subject_issuer)
                                         .public_key(key=private_key.public_key())
                                         .serial_number(number=x509.random_serial_number())
                                         .not_valid_before(time=valid_from)
                                         .not_valid_after(time=valid_to)
                                         .add_extension(extval=x509.BasicConstraints(ca=True,
                                                                                     path_length=None),
                                                        critical=True)
                                         .sign(private_key=private_key,
                                               algorithm=hashes.SHA256()))
    # format the output
    result: tuple[str, str] | tuple[bytes, bytes] | tuple[x509.Certificate, RSAPrivateKey]
    if fmt == "der":
        result = (certificate.public_bytes(encoding=Encoding.DER),
                  private_key.private_bytes(encoding=Encoding.DER,
                                            format=PrivateFormat.TraditionalOpenSSL,
                                            encryption_algorithm=NoEncryption()))
    elif fmt == "pem":
        result = (certificate.public_bytes(encoding=Encoding.PEM).decode(encoding="utf-8"),
                  private_key.private_bytes(encoding=Encoding.PEM,
                                            format=PrivateFormat.TraditionalOpenSSL,
                                            encryption_algorithm=NoEncryption()).decode(encoding="utf-8"))
    else:
        result = (certificate, private_key)

    return result


def cert_make_pfx(common_name: str,
                  organization: str,
                  locality: str,
                  province: str,
                  country: str,
                  valid_from: datetime = None,
                  valid_for: int = 365,
                  pfx_filepath: Path | str = None,
                  pfx_password: str | bytes = None,
                  friendly_name: str | bytes = None,
                  cert_set: list[x509.Certificate] = None) -> bytes:
    """
    Generate a PFX (PKCS#12) file by way of a self-signed x509 digital certificate/private key set.

    If specified, *pfx_filepath* must contain a valid filepath. If provided without a file extension,
    the latter is set to *.pfx". If the file already exists, it will be overwritten.

    :param common_name: the certificates's common name, typically the holder's identification
    :param organization: the reference organization
    :param locality: the reference locality, typically the name of a city
    :param province: the reference province, typically the name of a state or province
    :param country: the two-letter *Country Code Top-Level Domain* (ccTLD)
    :param valid_from: the certificate's validation start (defaults to now)
    :param valid_for: the certificate's validation period, in days (defaults to one year)
    :param pfx_filepath: optional path to output the *.pfx* file (no file written, if not provided)
    :param pfx_password: optional password for the *.pfx* file (file not encrypted, if not provided)
    :param friendly_name: optional friendly name for the certificate in the *.pfx* file
    :param cert_set: optional set of certificates to include in the *.pfx* file
    :return: the *.pfx* file bytes
    """
    # obtain a self-signed certificate and its corresponding private key
    pkcs7_data: tuple[x509.Certificate, RSAPrivateKey] = cert_make_certificate(common_name=common_name,
                                                                               organization=organization,
                                                                               locality=locality,
                                                                               province=province,
                                                                               country=country,
                                                                               valid_from=valid_from,
                                                                               valid_for=valid_for,
                                                                               fmt="x509")
    certificate: x509.Certificate = pkcs7_data[0]
    private_key: RSAPrivateKey = pkcs7_data[1]

    # obtain and return the PKCS#12 structure
    return cert_build_pfx(certificate=certificate,
                          private_key=private_key,
                          pfx_filepath=pfx_filepath,
                          pfx_password=pfx_password,
                          friendly_name=friendly_name,
                          cert_set=cert_set)


def cert_build_pfx(certificate: x509.Certificate,
                   private_key: RSAPrivateKey,
                   pfx_filepath: Path | str = None,
                   pfx_password: str | bytes = None,
                   friendly_name: str | bytes = None,
                   cert_set: list[x509.Certificate] = None) -> bytes:
    """
    Build a PFX (PKCS#12) file from cryptography's *Certificate* and *RSAPrivateKey* objects.

    If specified, *pfx_filepath* must contain a valid filepath. If provided without a file extension,
    the latter is set to *.pfx". If the file already exists, it will be overwritten.

    :param certificate: the *cryptography.x509.Certificate* object
    :param private_key: the *cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey* object
    :param pfx_filepath: optional path to output the *.pfx* file (no file written, if not provided)
    :param pfx_password: optional password for the *.pfx* file (file not emcrypted, if not provided)
    :param friendly_name: optional friendly name for the certificate in the *.pfx* file
    :param cert_set: optional set of certificates to include in the *.pfx* file
    :return: The content of the *.pfx* file as bytes
    """
    if isinstance(friendly_name, str):
        friendly_name = friendly_name.encode(encoding="utf-8")
    encryption: KeySerializationEncryption
    if pfx_password is None:
        encryption = NoEncryption()
    else:
        if isinstance(pfx_password, str):
            pfx_password = pfx_password.encode(encoding="utf-8")
        encryption = BestAvailableEncryption(password=pfx_password)

    result: bytes = serialize_key_and_certificates(name=friendly_name,
                                                   key=private_key,
                                                   cert=certificate,
                                                   cas=cert_set,
                                                   encryption_algorithm=encryption)
    # write the PKCS#12 structure to a file
    if pfx_filepath:
        # make sure 'pfx_filepath' is a 'Path'
        pfx_filepath = Path(pfx_filepath)
        if not pfx_filepath.suffix:
            pfx_filepath = pfx_filepath.with_suffix(".pfx")
        # write the file
        with pfx_filepath.open("wb") as f:
            f.write(result)

    return result


def cert_bundle_certs(certs: list[str | bytes | x509.Certificate],
                      fmt: Literal["der", "pem"] = "der") -> str | bytes:
    """
    Bundle the certificates in *certs* into a single char string or byte string.

    The format of the individual input certificates can be:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*
        - *x509*: *cryptography.x509.Certificate*, object as non-serialized binary

    *PEM* is a text format with delimiters (*-----BEGIN CERTIFICATE-----* and *-----END CERTIFICATE-----*),
    whereas *DER* is a binary format having no delimiters. When bundling multiple *DER* certificates into a
    single byte string, a way to delimit them for later unbundling is therefore needed. The approach used
    here is to prefix each certificate with a fixed-size header holding its length.

    The output format of the bundle is given by *fmt*:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*

    :param certs: the certificates to bundle
    :param fmt: the output format to use for the individual certificates (defaults to *der*)
    :return: the bundle certificatrs as a char string or byte string
    """
    # initialize the return variable
    result: str | bytes = "" if fmt == "pem" else b""

    for cert in certs:
        if isinstance(cert, x509.Certificate):
            # 'cert' is a certificate object
            if fmt == "pem":
                result += cert.public_bytes(encoding=Encoding.PEM).decode(encoding="utf-8")
            else:
                cert_der = cert.public_bytes(encoding=Encoding.DER)
                result += struct.pack(">I", len(cert_der)) + cert_der
        elif isinstance(cert, str):
            # 'cert' is PEM-encoded
            if fmt == "pem":
                result += cert
            else:
                result += cert.encode(encoding="utf-8")
        # 'cert' is DER-encoded
        elif fmt == "pem":
            result += (struct.pack(">I", len(cert)) + cert).decode(encoding="utf-8")
        else:
            result += struct.pack(">I", len(cert)) + cert

    return result


def cert_unbundle_certs(certs: str | bytes,
                        fmt: Literal["der", "pem", "x509"] = "der") -> list[str | bytes | x509.Certificate]:
    """
    Unbundle the certificates in *certs* into a list of certificates.

    The format of the individual input certificates in the bundle can be:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*

    *PEM* is a text format with delimiters (*-----BEGIN CERTIFICATE-----* and *-----END CERTIFICATE-----*),
    whereas *DER* is a binary format having no delimiters. When bundling multiple *DER* certificates into a
    single byte string, a way to delimit them for later unbundling was therefore needed. The approach expected
    here is to have each certificate prefixed with a fixed-size header holding its length.

    The return format of the individual certificates is specified in *fmt*:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*
        - *x509*: *cryptography.x509.Certificate*, object as non-serialized binary

    :param certs: the certificates to bundle
    :param fmt: the output format to use for the individual certificates (defaults to *der*)
    :return: the bundle certificatrs as a char string or byte string
    """
    # initialize the return variable
    result: list[str | bytes | x509.Certificate] = []

    # iterate on the certificates
    if isinstance(certs, str):
        # 'certs' is a PEM bundle
        certs_pem: list[str] = certs.split("-----END CERTIFICATE-----")
        for cert_pem in certs_pem:
            cert_pem += "-----END CERTIFICATE-----"
            if fmt == "pem":
                result.append(cert_pem)
            else:
                cert_x509: x509.Certificate = x509.load_pem_x509_certificate(data=cert_pem.encode("utf-8"))
                if cert_pem == "der":
                    result.append(cert_x509.public_bytes(encoding=Encoding.DER))
                else:
                    result.append(cert_x509)
    else:
        # 'certs' is a DER bundle
        offset: int = 0
        while offset < len(certs):
            # obtain the certificate length as encoded in the 4-byte header
            length: int = struct.unpack(">I", certs[offset:offset+4])[0]
            offset += 4
            # extract the certificate bytes
            cert_der: bytes = certs[offset:offset+length]
            offset += length
            if fmt == "der":
                result.append(cert_der)
            else:
                # load certificate
                cert_x509: x509.Certificate = x509.load_der_x509_certificate(data=cert_der)
                if fmt == "pem":
                    result.append(cert_x509.public_bytes(encoding=Encoding.PEM).decode(encoding="utf-8"))
                else:
                    result.append(cert_x509)
    return result


def cert_get_trusted(fmt: Literal["der", "pem", "x509"]) -> list[str | bytes | x509.Certificate]:
    """
    Retrieve the certificates in the trusted store of the host OS.

    The return format of the individual certificates is specified in *fmt*:
        - *der*: *Distinguished Encoding Rules*, binary as *bytes* (the default)
        - *pem*: *Privacy-Enhanced Mail*, text-based as *str*
        - *x509*: *cryptography.x509.Certificate*, object as non-serialized binary

    :param fmt: the output format to use for the individual certificates (defaults to *der*)
    :return: the list of certificates in the trusted store of the host OS
    """
    # initialize the return variable
    result: list[str | bytes | x509.Certificate] = []

    # read the trusted store
    ca_path: Path = Path(certifi.where())
    with ca_path.open("rb") as f:
        pem_data: bytes = f.read()

    # iterate on the certificates
    certs_pem: list[bytes] = pem_data.split(b"-----END CERTIFICATE-----")
    for cert_pem in certs_pem:
        cert_pem += b"-----END CERTIFICATE-----"
        if fmt == "pem":
            result.append(cert_pem.decode(encoding="utf-8"))
        else:
            cert_x509: x509.Certificate = x509.load_pem_x509_certificate(data=cert_pem)
            if cert_pem == "der":
                result.append(cert_x509.public_bytes(encoding=Encoding.DER))
            else:
                result.append(cert_x509)

    return result


def cert_verify_chain(cert_chain: list[x509.Certificate],
                      trusted_roots: list[x509.Certificate] = None,
                      errors: list[str] = None,
                      logger: Logger = None) -> bool:
    """
    Validate the certificates *cert_chain*, optionally using the trusted roots in *trusted_roots*.

    The verification is interrupted once the first problem is found.

    :param cert_chain: the certificate chain to validate
    :param trusted_roots: optional list of trusted roots to check the last certificate with
    :param errors: incidental errors (may be non-empty)
    :param logger: optional logger
    :return: True if *cert_chain* is valid, *False* otherwise
    """
    # define a local errors lista
    curr_errors: list[str] = []

    # check validity and BasicConstraints
    now: datetime = datetime.now(tz=UTC)
    err_msg: str | None = None
    for idx, cert in enumerate(iterable=cert_chain):
        if now < cert.not_valid_before:
            err_msg = f"Certificate '{cert.subject}' not yet valid"
        elif now > cert.not_valid_after:
            err_msg = f"Certificate '{cert.subject}' expired"
        elif idx > 0:  # intermediates
            bc: x509.BasicConstraints = cert.extensions.get_extension_for_class(x509.BasicConstraints).value
            if not bc.ca:
                err_msg = f"'{cert.subject}' is not a CA"
            elif isinstance(bc.path_length, int) and len(cert_chain) - idx - 1 > bc.path_length:
                err_msg = f"Path length constraint violated for '{cert.subject}'"
        if err_msg:
            break

    if not err_msg:
        # verify signatures
        for i in range(len(cert_chain) - 1):
            cert_verify_signature(cert=cert_chain[i],
                                  issuer=cert_chain[i + 1],
                                  errors=curr_errors,
                                  logger=logger)
            if curr_errors:
                break

        if not curr_errors and trusted_roots:
            # check last cert against trusted roots
            last_cert = cert_chain[-1]
            if not any(last_cert.subject == root.subject for root in trusted_roots):
                err_msg = "Chain does not terminate in a trusted root"

        # revocation checks
        if not err_msg:
            for idx, cert in enumerate(cert_chain[:-1]):  # leaf and intermediates
                issuer = cert_chain[idx + 1]
                if not cert_verify_revocation(cert=cert,
                                              issuer=issuer,
                                              logger=logger):
                    curr_errors.append(f"Certificate '{cert.subject}' has been revoked")
                    break
    if err_msg:
        if logger:
            logger.error(msg=err_msg)
        curr_errors.append(err_msg)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return not curr_errors


def cert_verify_signature(cert: x509.Certificate,
                          issuer: x509.Certificate,
                          errors: list[str] = None,
                          logger: Logger = None) -> bool:
    """
    Verify whether *cert*'s signature is valid.

    :param cert: the reference certificate
    :param issuer: the certificater issuer
    :param errors: incidental errors (may be non-empty)
    :param logger: optional logger
    :return: *True* if the signature is valid, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the certificate's public key
    public_key: PublicKeyTypes = issuer.public_key()

    # verify the signature
    try:
        if isinstance(public_key, RSAPublicKey):
            # determine the signature padding used
            sig_oid: str = cert.signature_algorithm_oid.dotted_string
            if sig_oid == "1.2.840.113549.1.1.10":  # RSASSA-PSS
                chosen_padding: padding.AsymmetricPadding = padding.PSS(
                    mgf=padding.MGF1(cert.signature_hash_algorithm),
                    salt_length=padding.PSS.MAX_LENGTH
                )
            else:
                chosen_padding: padding.AsymmetricPadding = padding.PKCS1v15()

            public_key.verify(cert.signature,
                              cert.tbs_certificate_bytes,
                              chosen_padding,
                              cert.signature_hash_algorithm)
        else:
            public_key.verify(cert.signature,
                              cert.tbs_certificate_bytes,
                              cert.signature_hash_algorithm)
        result = True
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=exc_err)
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def cert_verify_revocation(cert: x509.Certificate,
                           issuer: x509.Certificate,
                           logger: Logger = None) -> bool:
    """
    Verify whether *cert* is in good standing, that is, it has not been revoked.

    Two attempts are carried out to make sure ther certificate is still in good standing:
        - the appropriate *Certificate Revocation Lists* (CRLs) are inspected
        - the newer *Online Certificate Status Protocol* (OCSP) protocol is used

    :param cert: the reference certificate
    :param issuer: the certificate issuer
    :param logger: optional logger
    :return: *True* if the certificate is in good standing, *False* otherwise
    """
    # initialize the return variable
    result: bool = True

    # retrieve the CRL verification address
    crl_ext_value: x509.extensions.ExtensionTypeVar | None = None
    with suppress(x509.ExtensionNotFound):
        crl_ext: x509.extensions.Extension = cert.extensions.get_extension_for_class(
            extclass=x509.CRLDistributionPoints
        )
        crl_ext_value = crl_ext.value

    # check CRL distribution points
    for dp in crl_ext_value or []:
        for uri in dp.full_name:
            url: str = uri.value
            if url.startswith("http"):
                if logger:
                    logger.debug(msg=f"GET {url}")
                reply: requests.Response = requests.get(url=url,
                                                        timeout=None)
                if 200 >= reply.status_code <= 300:
                    if logger:
                        logger.debug("GET success")
                    crl_data: bytes = reply.content
                    crl: x509.CertificateRevocationList = x509.load_der_x509_crl(data=crl_data)
                    for revoked in crl:
                        if revoked.serial_number == cert.serial_number:
                            result = False
                            if logger:
                                logger.error(f"Certificate {cert.subject} has been revoked")
                elif logger:
                    logger.warning(msg=f"GET failure, status {reply.status_code}")

    # use OCSP protocol for further checking
    if result:
        aia_value: x509.extensions.ExtensionTypeVar | None = None
        with suppress(x509.ExtensionNotFound):
            aia: x509.extensions.Extension = cert.extensions.get_extension_for_class(
                extclass=x509.AuthorityInformationAccess
            )
            aia_value = aia.value

        ocsp_urls: list[Any] = [desc.access_location.value for desc in aia_value
                                if desc.access_method.dotted_string == "1.3.6.1.5.5.7.48.1"]
        for url in ocsp_urls:
            print(f"Performing OCSP check at {url}")
            builder: OCSPRequestBuilder = OCSPRequestBuilder()
            # ruff: noqa: S303
            builder = builder.add_certificate(cert=cert,
                                              issuer=issuer,
                                              algorithm=hashes.SHA1())
            ocsp_request: x509.ocsp.OCSPRequest = builder.build()
            headers = {"Content-Type": "application/ocsp-request",
                       "Accept": "application/ocsp-response"}
            if logger:
                logger.debug(msg=f"GET {url}")
            reply: requests.Response = requests.post(url=url,
                                                     data=ocsp_request.public_bytes(encoding=Encoding.DER),
                                                     headers=headers,
                                                     timeout=None)
            if 200 >= reply.status_code <= 300:
                if logger:
                    logger.debug("GET success")
                ocsp_resp = load_der_ocsp_response(data=reply.content)

                if ocsp_resp.response_status.name == "successful":
                    cert_status = ocsp_resp.certificate_status.name
                    if cert_status == "revoked":
                        result = False
                        if logger:
                            logger.error(msg=f"Certificate '{cert.subject}' has been revoked")
                    elif logger:
                        logger.debug(msg=f"Certificate '{cert.subject}' status '{cert_status}'")
                elif logger:
                    logger.warning(msg=f"OCSP responder returned '{ocsp_resp.response_status.name}'")
            elif logger:
                logger.warning(msg=f"GET failure, status {reply.status_code}")

    return result



from __future__ import annotations
import base64
import sys
import tempfile
from asn1crypto import cms, tsp
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from logging import Logger
from pathlib import Path
from pyhanko.pdf_utils.generic import (
    ArrayObject as PhArrayObject,
    DictionaryObject as PhDictionaryObject,
    IndirectObject as PhIndirectObject,
    NameObject as PhNameObject
)
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign.signers import PdfSigner, SimpleSigner, PdfSignatureMetadata
from pyhanko.sign.fields import SigFieldSpec
from pyhanko.stamp import TextStampStyle
from pyhanko.sign.timestamps import HTTPTimeStamper
from PyPDF2 import PdfReader
from PyPDF2.generic import ArrayObject, ByteStringObject, DictionaryObject, Field
from pypomes_core import TZ_LOCAL, file_get_data, exc_format
from requests.auth import HTTPBasicAuth
from typing import Any, Literal

from .crypto_common import HashAlgorithm, ChpHash, _chp_hash


class CryptoPdf:
    """
    Python code to extract crypto data from a *PAdES* compliant, digitally signed, PDF file.

    The crypto data is mostly in *Cryptographic Message Syntax* (CMS), a standard for digitally signing,
    digesting, authenticating, and encrypting arbitrary message content. In the case of the *PAdES* standard,
    some deviations exist, due to the utilization of PDF dictionaries to hold some of the data.

    These are the instance variables:
        - signatures: list of *SignatureInfo*, holds the crypto data of the document's signatures
        - pdf_bytes: holds the full bytes content of the PDF file, on which the payload ranges are applied
    """
    # class-level logger
    logger: Logger | None = None

    @dataclass(frozen=True)
    class SignatureInfo:
        """
        These are the attributes holding the signature data.
        """
        payload_range: tuple[int, int, int, int]  # the range of bytes comprising the payload
        payload_hash: bytes                       # the payload hash
        hash_algorithm: HashAlgorithm             # the algorithm used to calculate the payload hash
        signature: bytes                          # the digital signature
        signature_algorithm: str                  # the algorithm used to generate the signature
        signature_timestamp: datetime             # the signature's timestamp
        public_key: PublicKeyTypes                # the public key (most likely, RSAPublicKey)
        signer_common_name: str                   # the name of the certificate's signer
        signer_cert: x509.Certificate             # the reference certificate (latest one in the chain)
        cert_serial_number: int                   # the certificate's serial nmumber
        cert_chain: list[bytes]                   # the serialized X509 certificate chain (in DER format)

        # TSA (Time Stamping Authority) data
        tsa_timestamp: datetime                   # the signature's timestamp
        tsa_policy: str                           # the TSA's policy
        tsa_serial_number: str                    # the timestamping's serial number

    def __init__(self,
                 pdf_data: Path | str | bytes,
                 pdf_pwd: str = None,
                 errors: list[str] = None) -> None:
        """
        Instantiate the *CryptoPdf* class, and extract the relevant crypto data.

        The nature of *pdf_data* depends on its data type:
          - type *bytes*: *pdf_data* holds the data (used as is)
          - type *str*: *pdf_data* holds the data (used as utf8-encoded)
          - type *Path*: *pdf_data* is a path to a file holding the data

        If *pdf_data* is encrypted, the decryption password must be provided in *pdf_pwd*.

        :param pdf_data: a digitally signed, *PAdES* cxonformant, PDF file
        :param pdf_pwd: optional password for file decryption
        :param errors: incidental errors (may be non-empty)
        """
        # declare/initialize the instance variables
        self.signatures: list[CryptoPdf.SignatureInfo] = []
        self.pdf_bytes: bytes

        # retrieve the PDF data
        self.pdf_bytes = file_get_data(file_data=pdf_data)

        # define a local errors list
        curr_errors: list[str] = []

        pdf_stream: BytesIO = BytesIO(initial_bytes=self.pdf_bytes)
        pdf_stream.seek(0)

        # retrieve the signature fields
        reader: PdfReader = PdfReader(stream=pdf_stream,
                                      password=pdf_pwd)
        sig_fields: list[Field] = [field for field in reader.get_fields().values()
                                   if field.get("/FT") == "/Sig"] or []

        # process the signature fields
        for sig_field in sig_fields:
            sig_dict: DictionaryObject = sig_field.get("/V")
            contents: ByteStringObject = sig_dict.get("/Contents")

            # extract the payload
            byte_range: ArrayObject = sig_dict.get("/ByteRange")
            from_1, len_1, from_2, len_2 = byte_range
            payload_range: tuple[int, int, int, int] = (int(from_1), int(len_1), int(from_2), int(len_2))
            payload: bytes = self.pdf_bytes[from_1:from_1+len_1] + self.pdf_bytes[from_2:from_2+len_2]

            # extract signature data (CMS structure)
            sig_obj: ByteStringObject = contents.get_object()
            cms_obj: cms.ContentInfo = cms.ContentInfo.load(encoded_data=sig_obj)
            signed_data: cms.SignedData = cms_obj["content"]
            signer_info: cms.SignerInfo = signed_data["signer_infos"][0]

            # extract the signature and its algorithms
            signature: bytes = signer_info["signature"].native
            signature_algorithm: str = signer_info["signature_algorithm"]["algorithm"].native
            alg_name: str = signer_info["digest_algorithm"]["algorithm"].native
            hash_algorithm: HashAlgorithm = HashAlgorithm(alg_name)

            # extract signature timestamp and payload hash
            chp_hash: ChpHash = _chp_hash(alg=hash_algorithm)
            payload_hash: bytes | None = None
            signature_timestamp: datetime | None = None
            if "signed_attrs" in signer_info:
                signed_attrs: cms.CMSAttributes = signer_info["signed_attrs"]
                for signed_attr in signed_attrs:
                    attr_type: str = signed_attr["type"].native
                    match attr_type:
                        case "message_digest":
                            payload_hash = signed_attr["values"][0].native
                        case "signing_time":
                            signature_timestamp = signed_attr["values"][0].native
            # validate hash
            from .crypto_pomes import crypto_hash
            effective_hash: bytes = crypto_hash(msg=payload,
                                                alg=hash_algorithm)
            if not payload_hash:
                payload_hash = effective_hash
            elif payload_hash != effective_hash:
                msg: str = f"Invalid digest for signature timestamp '{signature_timestamp}'"
                if CryptoPdf.logger:
                    CryptoPdf.logger.error(msg=msg)
                curr_errors.append(msg)
                break

            # build the certificate chain
            cert_chain: list[bytes] = []
            certs: cms.CertificateSet = signed_data["certificates"]
            for cert in certs:
                der_bytes: bytes = cert.dump()
                cert_chain.append(der_bytes)

            # extract certificates and public key
            signer_cert: x509.Certificate = x509.load_der_x509_certificate(data=cert_chain[0])
            public_key: PublicKeyTypes = signer_cert.public_key()
            cert_serial_number: int = signer_cert.serial_number

            # identify signer
            subject: x509.name.Name = signer_cert.subject
            signer_common_name: str = subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value

            # TSA timestamp info (optional)
            tsa_timestamp: datetime | None = None
            tsa_policy: str | None = None
            tsa_serial_number: str | None = None

            unsigned_attrs: cms.CMSAttributes = signer_info["unsigned_attrs"]
            for unsigned_attr in unsigned_attrs:
                attr_type: str = unsigned_attr["type"].native
                if attr_type == "signature_time_stamp_token":
                    try:
                        # the timestamp token is a CMS SignedData structure -'dump()' gets the raw DER bytes
                        attr_values: cms.SetOfContentInfo = unsigned_attr["values"]
                        timestamp_token: cms.ContentInfo = cms.ContentInfo.load(attr_values[0].dump())

                        # extract the TSTInfo structure
                        tst_signed_data: cms.SignedData = timestamp_token["content"]
                        tst_info: tsp.TSTInfo = tst_signed_data["encap_content_info"]["content"].parsed

                        # extract TSA timestamp details
                        tsa_timestamp = tst_info["gen_time"].native
                        tsa_policy = tst_info["policy"].native
                        tsa_serial_number = hex(tst_info["serial_number"].native)

                    except Exception as e:
                        # unable to obtain TAS data: error parsing token
                        if CryptoPdf.logger:
                            msg: str = exc_format(exc=e,
                                                  exc_info=sys.exc_info())
                            CryptoPdf.logger.error(msg=msg)
                    break

            # verify the signature
            try:
                if isinstance(public_key, RSAPublicKey):
                    # determine the signature padding used
                    if "pss" in signature_algorithm:
                        signature_padding: padding.AsymmetricPadding = padding.PSS(
                            mgf=padding.MGF1(algorithm=chp_hash),
                            salt_length=padding.PSS.MAX_LENGTH
                        )
                    else:
                        signature_padding: padding.AsymmetricPadding = padding.PKCS1v15()
                    public_key.verify(signature=signature,
                                      data=payload_hash,
                                      padding=signature_padding,
                                      algorithm=Prehashed(chp_hash))
                else:
                    public_key.verify(signature=signature,
                                      data=payload_hash,
                                      algorithm=Prehashed(chp_hash))
            except Exception as e:
                if CryptoPdf.logger:
                    msg: str = exc_format(exc=e,
                                          exc_info=sys.exc_info()) + f" signed by {signer_common_name}"
                    CryptoPdf.logger.warning(msg=msg)

            # build the signature's crypto data and save it
            sig_info: CryptoPdf.SignatureInfo = CryptoPdf.SignatureInfo(
                payload_range=payload_range,
                payload_hash=payload_hash,
                hash_algorithm=hash_algorithm,
                signature=signature,
                signature_algorithm=signature_algorithm,
                signature_timestamp=signature_timestamp,
                public_key=public_key,
                signer_common_name=signer_common_name,
                signer_cert=signer_cert,
                cert_serial_number=cert_serial_number,
                cert_chain=cert_chain,
                tsa_timestamp=tsa_timestamp,
                tsa_policy=tsa_policy,
                tsa_serial_number=tsa_serial_number
            )
            self.signatures.append(sig_info)

        if not curr_errors and not self.signatures:
            msg: str = "No digital signatures found in PDF file"
            if CryptoPdf.logger:
                CryptoPdf.logger.error(msg=msg)
            curr_errors.append(msg)

        if curr_errors and isinstance(errors, list):
            errors.extend(curr_errors)

    def get_digest(self,
                   fmt: Literal["base64", "bytes"],
                   sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the digest associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the digest, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPdf.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.payload_hash \
            if fmt == "bytes" else base64.b64encode(s=sig_info.payload_hash).decode(encoding="utf-8")

    def get_signature(self,
                      fmt: Literal["base64", "bytes"],
                      sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the signature associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the signature, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPdf.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.signature \
            if fmt == "bytes" else base64.b64encode(s=sig_info.signature).decode(encoding="utf-8")

    def get_public_key(self,
                       fmt: Literal["base64", "der", "pem"],
                       sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the public key associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        These are the supported formats:
            - *der*: the raw binary representation of the key
            - *pem*: the Base64-encoded key with headers and line breaks
            - *base64*: the Base64-encoded DER bytes

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the public key, as per *fmt* (*str* or *bytes*)
        """
        # declare the return variable
        result: str | bytes

        sig_info: CryptoPdf.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        if fmt == "pem":
            result = sig_info.public_key.public_bytes(encoding=Encoding.PEM,
                                                      format=PublicFormat.SubjectPublicKeyInfo)
            result = result.decode(encoding="utf-8")
        else:
            result = sig_info.public_key.public_bytes(encoding=Encoding.DER,
                                                      format=PublicFormat.SubjectPublicKeyInfo)
            if fmt == "base64":
                result = base64.b64encode(s=result).decode(encoding="utf-8")

        return result

    def get_cert_chain(self,
                       sig_seq: int = 0) -> list[bytes]:
        """
        Retrieve the certificate chain associated with a reference signature, as specified in *sig_seq*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the signature, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPdf.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.cert_chain

    def get_metadata(self,
                     sig_seq: int = 0) -> dict[str, Any]:
        """
        Retrieve the certificate chain metadata associated with a reference signature, as specified in *sig_seq*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the certificate chain metadata associated with the reference signature
        """
        # declare the return variable
        result: dict[str, Any]

        sig_info: CryptoPdf.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        cert: x509.Certificate = sig_info.signer_cert

        result: dict[str, Any] = {
            "signer-common-name": sig_info.signer_common_name,
            "hash-algorithm": sig_info.hash_algorithm,
            "signature-algorithm": sig_info.signature_algorithm,
            "signature-timestamp": sig_info.signature_timestamp,
            "cert-serial-number": sig_info.cert_serial_number,
            "cert-not-before": cert.not_valid_before,
            "cert-not-after": cert.not_valid_after,
            "cert-subject": cert.subject.rfc4514_string(),
            "cert-issuer": cert.issuer.rfc4514_string(),
            "cert-chain-length": len(sig_info.cert_chain)
        }
        # add the TSA details
        if sig_info.tsa_serial_number:
            result.update({
                "tsa-timestamp": sig_info.tsa_timestamp,
                "tsa-policy": sig_info.tsa_policy,
                "tsa-serial-number": sig_info.tsa_serial_number
            })

        return result

    def __get_sig_info(self,
                       sig_seq: int) -> CryptoPdf.SignatureInfo:
        """
        Retrieve the signature metadata of a reference signature, as specified in *sig_seq*.

        The natural ordering of the signatures in a *PAdES* compliant, digitally signed, PDF file is the
        chronological *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position
        of the last signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of the range
        of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the reference signature's metadata

        """
        sig_ordinal: int = max(-1, len(self.signatures) - sig_seq - 1)
        return self.signatures[sig_ordinal]

    @staticmethod
    def set_logger(logger: Logger) -> None:
        """
        Configure the logger to be used in this module's operations.

        :param logger: the operations logger
        """
        CryptoPdf.logger = logger

    @staticmethod
    def create(pdf_in: Path | str | bytes,
               pfx_data: Path | str | bytes,
               pfx_pwd: str | bytes = None,
               pdf_out: Path | str = None,
               location: str = None,
               reason: str = None,
               make_visible: bool = False,
               page_num: int = 0,
               box: tuple[int, int, int, int] = (50, 50, 250, 150),
               tsa_url: str = None,
               tsa_username: str = None,
               tsa_password: str = None,
               errors: list[str] = None) -> CryptoPdf:
        """
        Digitally sign a PDF file in *PAdES* format using an A1 certificate.

        The natures of *pdf_in* and *pfx_data* depend on their respective data types:
          - type *bytes*: holds the data (used as is)
          - type *str*: holds the data (used as utf8-encoded)
          - type *Path*: is a path to a file holding the data

        Supports visible signature appearance, TSA timestamping, and multiple signatures.

        :param pdf_in: input PDF data
        :param pfx_data: the PKCS#12 (*.pfx*) data, containing A1 certificate and private key
        :param pfx_pwd: password for the *.pfx* data (if not provided, *pfx_data* is assumed to be unencrypted)
        :param pdf_out: path to output the signed/re-signed PDF file (optional, no output if not provided)
        :param location: location of signing
        :param reason: reason for signing
        :param make_visible: whether to include a visible signature appearance
        :param page_num: page number for the visible signature (0-based, specify -1 for last page)
        :param box: (x1, y1, x2, y2) coordinates for the signature box
        :param tsa_url: TSA server URL for timestamping
        :param tsa_username: TSA username (if required)
        :param tsa_password: TSA password (if required)
        :param errors: incidental errors (may be non-empty)
        :return: the corresponding instance of *CryptoPdf*, or *None* if error
        """
        # initialize the return variable
        result: CryptoPdf | None = None

        # load the signing certificate and key from a PKCS#12 file
        is_temp: bool = False
        pwd_bytes = pfx_pwd.encode() if isinstance(pfx_pwd, str) else pfx_pwd
        if isinstance(pfx_data, str):
            pfx_data = pfx_data.encode(encoding="utf-8")
        if isinstance(pfx_data, bytes):
            # PyHanko's SimpleSigner requires a path to a file
            is_temp = True
            with tempfile.NamedTemporaryFile(mode="wb",
                                             delete=False) as tmp:
                tmp.write(pfx_data)
                pfx_data = Path(tmp.name)
        simple_signer: SimpleSigner = SimpleSigner.load_pkcs12(pfx_file=pfx_data,
                                                               passphrase=pwd_bytes)
        if is_temp:
            pfx_data.unlink(missing_ok=True)
        if simple_signer:
            # cconfigure stamp style
            stamp_style: TextStampStyle | None = None
            if make_visible:
                stamp_text: str = f"Signed by {simple_signer.subject_name}"
                if reason:
                    stamp_text += f"\nReason: {reason}"
                if location:
                    stamp_text += f"\nLocation: {location}"
                stamp_text += f"\nDate: {datetime.now(tz=TZ_LOCAL)}"
                stamp_style = TextStampStyle(stamp_text=stamp_text)

            # configure TSA
            timestamper: HTTPTimeStamper | None = None
            if tsa_url:
                auth: HTTPBasicAuth | None = None
                if tsa_username and tsa_password:
                    auth = HTTPBasicAuth(username=tsa_username,
                                         password=tsa_password)
                timestamper = HTTPTimeStamper(url=tsa_url,
                                              auth=auth,
                                              timeout=None)

            # open PDF file for incremental signing
            pdf_in = file_get_data(file_data=pdf_in)
            pdf_stream: BytesIO = BytesIO(initial_bytes=pdf_in)
            output_buf: BytesIO | None = None

            pdf_stream.seek(0)
            writer: IncrementalPdfFileWriter = IncrementalPdfFileWriter(input_stream=pdf_stream,
                                                                        strict=False)
            # Use PdfFileReader to inspect fields
            reader: PdfFileReader = PdfFileReader(pdf_stream)
            acroform_ref: PhIndirectObject = reader.root.get("/AcroForm")
            sig_field: str | None = None
            field_count: int = 0

            if acroform_ref:
                # dereference IndirectObject
                acroform: PhDictionaryObject = acroform_ref.get_object()
                fields_array: PhArrayObject = acroform.get("/Fields") or []
                for field_ref in fields_array:
                    field_obj: PhDictionaryObject = field_ref.get_object()
                    field_type: PhNameObject = field_obj.get("/FT")
                    if field_type == "/Sig":
                        field_count += 1
                        field_value: PhIndirectObject = field_obj.get("/V")
                        if sig_field is None and field_value is None:
                            # use existing unused field
                            sig_field = field_obj.get("/T")
            if sig_field:
                # no need to create a new field
                new_field_spec = None
            else:
                # obtain the last page index
                if page_num == -1:
                    save_pos: int = pdf_stream.tell()
                    pdf_stream.seek(0)
                    temp_reader: PdfReader = PdfReader(stream=pdf_stream)
                    page_num = len(temp_reader.pages) - 1
                    del temp_reader
                    pdf_stream.seek(save_pos)
                # create a new field
                sig_field = f"Signature{field_count + 1}"
                new_field_spec = SigFieldSpec(sig_field_name=sig_field,
                                              box=box,
                                              on_page=page_num)
            # create signature metadata
            sig_metadata: PdfSignatureMetadata = PdfSignatureMetadata(field_name=sig_field,
                                                                      reason=reason,
                                                                      location=location)
            # create PdfSigner
            pdf_signer: PdfSigner = PdfSigner(signature_meta=sig_metadata,
                                              signer=simple_signer,
                                              timestamper=timestamper,
                                              stamp_style=stamp_style,
                                              new_field_spec=new_field_spec)
            try:
                output_buf = pdf_signer.sign_pdf(pdf_out=writer)
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if CryptoPdf.logger:
                    CryptoPdf.logger.error(msg=exc_err)
                if isinstance(errors, list):
                    errors.append(exc_err)

            # write the signed PDF
            if output_buf and pdf_out:
                # make sure 'pdf_out' is a 'Path'
                pdf_out = Path(pdf_out)
                signed_pdf: bytes = output_buf.read()
                with pdf_out.open("wb") as out_f:
                    out_f.write(signed_pdf)
        else:
            msg: str = "Unable to load PKCS#12 data from 'pfx_data'"
            if CryptoPdf.logger:
                CryptoPdf.logger.error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)

        return result



from .cert_pomes import (
    cert_make_certificate, cert_make_pfx, cert_build_pfx,
    cert_get_trusted, cert_bundle_certs, cert_unbundle_certs,
    cert_verify_chain, cert_verify_signature, cert_verify_revocation
)
from .crypto_aes import (
    CRYPTO_DEFAULT_SYMMETRIC_MODE, SymmetricMode,
    crypto_aes_encrypt, crypto_aes_decrypt
)
from .crypto_common import (
    CRYPTO_DEFAULT_HASH_ALGORITHM,
    SignatureMode, SignatureType, HashAlgorithm
)
from .crypto_pdf import (
    CryptoPdf
)
from .crypto_pkcs7 import (
    CryptoPkcs7
)
from .crypto_pomes import (
    crypto_hash, crypto_generate_rsa_keys,
    crypto_verify_pdf, crypto_verify_p7s,
    crypto_encrypt, crypto_decrypt,
    crypto_pwd_encrypt, crypto_pwd_verify
)
from .jwt_pomes import (
    jwt_convert, jwt_validate,
    jwt_get_header, jwt_get_payload,
    jwt_get_claim, jwt_get_claims, jwt_get_public_key
)

__all__ = [
    # cert_pomes
    "cert_make_certificate", "cert_make_pfx", "cert_build_pfx",
    "cert_get_trusted", "cert_bundle_certs", "cert_unbundle_certs",
    "cert_verify_chain", "cert_verify_signature", "cert_verify_revocation",
    # crypto_aes
    "SymmetricMode",
    "crypto_aes_encrypt", "crypto_aes_decrypt",
    # crypto_common
    "CRYPTO_DEFAULT_SYMMETRIC_MODE",
    "SignatureMode", "SignatureType", "HashAlgorithm",
    # crypto_pdf
    "CryptoPdf",
    # crypto_pkcs7
    "CryptoPkcs7",
    # crypto_pomes
    "CRYPTO_DEFAULT_HASH_ALGORITHM",
    "crypto_hash", "crypto_generate_rsa_keys",
    "crypto_verify_pdf", "crypto_verify_p7s",
    "crypto_encrypt", "crypto_decrypt",
    "crypto_pwd_encrypt", "crypto_pwd_verify",
    # jwt_pomes
    "jwt_convert", "jwt_validate",
    "jwt_get_header", "jwt_get_payload",
    "jwt_get_claim", "jwt_get_claims", "jwt_get_public_key"
]

from importlib.metadata import version
__version__ = version("pypomes_crypto")
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())



import hashlib
import pickle
import sys
from asn1crypto import cms, pem, x509 as asn1crypto_x509
from collections.abc import Iterable
from contextlib import suppress
from Crypto.Cipher import AES
from Crypto.Cipher._mode_ecb import EcbMode
from Crypto.Util.Padding import pad, unpad
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from io import BytesIO
from logging import Logger
from passlib.hash import argon2
from pathlib import Path
from pyhanko.keys import load_certs_from_pemder_data
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign.validation import validate_pdf_signature
from pyhanko.sign.validation.pdf_embedded import EmbeddedPdfSignature
from pyhanko.sign.validation.status import PdfSignatureStatus
from pyhanko_certvalidator import ValidationContext
from pypomes_core import file_get_data, exc_format

from .crypto_common import (
    CRYPTO_DEFAULT_HASH_ALGORITHM,
    HashAlgorithm, ChpHash, _chp_hash
)


def crypto_hash(msg: Path | str | bytes,
                alg: HashAlgorithm | str = CRYPTO_DEFAULT_HASH_ALGORITHM) -> bytes:
    """
    Compute the hash of *msg*, using the algorithm specified in *alg*.

    The nature of *msg* dependes on its data type:
        - type *bytes*: *msg* holds the data (used as is)
        - type *str*: *msg* holds the data (used as utf8-encoded)
        - type *Path*: *msg* is a path to a file holding the data
        - other: *pickle*'s serialization of *msg* is used

    Supported algorithms:
      *md5*, *blake2b*, *blake2s*, *sha1*, *sha224*, *sha256*, *sha384*,
      *sha512*, *sha3_224*, *sha3_256*, *sha3_384*, *sha3_512*, *shake_128*, *shake_256*.

    :param msg: the message to calculate the hash for
    :param alg: the algorithm to use (defaults to an environment-defined value, or to 'sha256')
    :return: the hash value obtained, or *None* if the hash could not be computed
    """
    # initialize the return variable
    result: bytes | None = None

    if alg in HashAlgorithm:
        # instantiate the hasher (undeclared type is '_Hash')
        hasher = hashlib.new(name=alg)

        if isinstance(msg, bytes):
            # argument is type 'bytes'
            hasher.update(msg)
            result = hasher.digest()

        elif isinstance(msg, str):
            # argument is type 'str'
            hasher.update(msg.encode())
            result = hasher.digest()

        elif isinstance(msg, Path):
            # argument is a file path
            buf_size: int = 128 * 1024
            with msg.open(mode="rb") as f:
                file_bytes: bytes = f.read(buf_size)
                while file_bytes:
                    hasher.update(file_bytes)
                    file_bytes = f.read(buf_size)
            result = hasher.digest()

        else:
            # argument is unknown
            with suppress(Exception):
                data: bytes = pickle.dumps(obj=msg)
                if data:
                    hasher.update(data)
                    result = hasher.digest()
    return result


def crypto_generate_rsa_keys(key_size: int = 2048) -> tuple[bytes, bytes]:
    """
    Generate and return a matching pair of *RSA* private and public keys.

    :param key_size: the key size (defaults to 2048 bytes)
    :return: a matching key pair *(private, public)* of serialized RSA keys
    """
    # generate the private key
    priv_key: RSAPrivateKey = rsa.generate_private_key(public_exponent=65537,
                                                       key_size=key_size)
    result_priv: bytes = priv_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                format=serialization.PrivateFormat.PKCS8,
                                                encryption_algorithm=serialization.NoEncryption())
    # generate the matching public key
    pub_key: RSAPublicKey = priv_key.public_key()
    result_pub: bytes = pub_key.public_bytes(encoding=serialization.Encoding.PEM,
                                             format=serialization.PublicFormat.SubjectPublicKeyInfo)
    # return the key pair
    return result_priv, result_pub


def crypto_encrypt(plaintext: Path | str | bytes,
                   key: bytes,
                   errors: list[str] = None) -> bytes:
    """
    Symmetrically encrypt *plaintext* using the given *key*.

    The *ECB* (Electronic CodeBook) symmetric block cipher is used. This is the most basic but also
    the weakest mode of operation available. Its use should be restricted to non-critical messages,
    or otherwise should be combined with a stronger cipher.

    It should also be noted that this cipher does not provides guarantees over the *integrity* of the message
    (i.e., it does not allow the receiver to establish whether the *ciphertext* was modified in transit).
    For a top-of-the-line symmetric block cipher, providing quality message *confidentiality* and *integrity*,
    consider using *crypto_aes_encrypt()/crypto_aes_decrypt()* in this package.

    The nature of *plaintext* depends on its data type:
      - type *bytes*: *plaintext* holds the data (used as is)
      - type *str*: *plaintext* holds the data (used as utf8-encoded)
      - type *Path*: *plaintext* is a path to a file holding the data

    The mandatory *key* must be 16, 24, or 32 bytes long.

    :param plaintext: the message to encrypt
    :param key: the cryptographic key (byte length must be 16, 24 or 32)
    :param errors: incidental error messages (may be non-empty)
    :return: the encrypted message, or *None* if error
    """
    # initialize the return variable
    result: bytes | None = None

    # obtain the data for encryption
    plaindata: bytes = file_get_data(file_data=plaintext)

    # build the cipher
    cipher: EcbMode = AES.new(key=key,
                              mode=AES.MODE_ECB)
    # encrypt the data
    try:
        result = cipher.encrypt(plaintext=pad(data_to_pad=plaindata,
                                              block_size=AES.block_size))
    except Exception as e:
        if isinstance(errors, list):
            exc_error: str = exc_format(exc=e,
                                        exc_info=sys.exc_info())
            errors.append(exc_error)

    return result


def crypto_decrypt(ciphertext: Path | str | bytes,
                   key: bytes,
                   errors: list[str] = None) -> bytes:
    """
    Symmetrically decrypt *ciphertext* using the given *key*.

    It is assumed that the *ECB* (Electronic CodeBook) symmetric block cipher was used to generate *ciphertext*.
    This is the most basic but also the weakest mode of operation available. Its use should be restricted to
    non-critical messages, or otherwise should be combined with a stronger cipher.

    It should also be noted that this cipher does not provides guarantees over the *integrity* of the message
    (i.e., it does not allow the receiver to establish whether *ciphertext* was modified in transit).
    For a top-of-the-line symmetric block cipher, providing quality message *confidentiality* and *integrity*,
    consider using *crypto_aes_encrypt()/crypto_aes_decrypt()* in this package.

    The nature of *ciphertext* depends on its data type:
      - type *bytes*: *ciphertext* holds the data (used as is)
      - type *str*: *ciphertext* holds the data (used as utf8-encoded)
      - type *Path*: *ciphertext* is a path to a file holding the data

     The *key* must be the same one used to generate *ciphertext*.

    :param ciphertext: the message to decrypt
    :param key: the cryptographic key
    :param errors: incidental error messages (may be non-empty)
    :return: the decrypted message, or *None* if error
    """
    # initialize the return variable
    result: bytes | None = None

    # obtain the data for decryption
    cipherdata: bytes = file_get_data(file_data=ciphertext)

    # build the cipher
    cipher: AES = AES.new(key=key,
                          mode=AES.MODE_ECB)
    # decrypt the data
    try:
        # HAZARD: the misnamed parameter ('plaintext') is left unnamed
        plaindata: bytes = cipher.decrypt(cipherdata)
        result = unpad(padded_data=plaindata,
                       block_size=AES.block_size)
    except Exception as e:
        if isinstance(errors, list):
            exc_error: str = exc_format(exc=e,
                                        exc_info=sys.exc_info())
            errors.append(exc_error)

    return result


def crypto_pwd_encrypt(pwd: str,
                       salt: bytes,
                       errors: list[str] = None) -> str:
    """
    Encrypt a password given in *pwd*, using the provided *salt*, and return it.

    :param pwd: the password to encrypt
    :param salt: the salt value to use (must be at least 8 bytes long)
    :param errors: incidental error messages (may be non-empty)
    :return: the encrypted password, or *None* if error
    """
    # initialize the return variable
    result: str | None = None

    try:
        pwd_hash: str = argon2.using(salt=salt).hash(secret=pwd)
        result = pwd_hash[pwd_hash.rfind("$")+1:]
    except Exception as e:
        if isinstance(errors, list):
            exc_error: str = exc_format(exc=e,
                                        exc_info=sys.exc_info())
            errors.append(exc_error)

    return result


def crypto_pwd_verify(plain_pwd: str,
                      cipher_pwd: str,
                      salt: bytes,
                      errors: list[str] = None) -> bool:
    """
    Verify, using the provided *salt*, whether the plaintext and encrypted passwords match.

    :param plain_pwd: the plaintext password
    :param cipher_pwd: the encryped password to verify
    :param salt: the salt value to use (must be at least 8 bytes long)
    :param errors: incidental error messages (may be non-empty)
    :return: *True* if they match, *False* otherwise
    """
    pwd_hash: str = crypto_pwd_encrypt(pwd=plain_pwd,
                                       salt=salt,
                                       errors=errors)
    return isinstance(pwd_hash, str) and cipher_pwd == pwd_hash


def crypto_verify_p7s(p7s_data: Path | str | bytes,
                      doc_data: Path | str | bytes = None,
                      errors: list[str] = None,
                      logger: Logger = None) -> bool | None:
    """
    Verify a PKCS#7 signature against a document.

    The natures of *p7s_data* and *doc_data* depend on their respective data types:
      - type *bytes*: holds the data (used as is)
      - type *str*: holds the data (used as utf8-encoded)
      - type *Path*: is a path to a file holding the data

    Both attached and detached signatures are properly handled, and full cryptographic verification
    (digest + RSA signature check) is performed. The PKCS#7 data provided in *p7s_data* contains the
    A1 certificate and its corresponding public key, the certificate chain, the original payload
    (if *attached* mode, only), and the digital signature.

    :param p7s_data: the PKCS#7 signature data containing A1 certificate, in *DER* or *PEM* format
    :param doc_data: the original document data (required for detached mode)
    :param errors: incidental errors (may be non-empty)
    :param logger: optional logger
     :return: *True* if signature is valid, *False* otherwise, or *None* if error
    """
    # initialize the return variable
    result: bool | None = None

    # retrieve the certificate raw bytes (if PEM, convert to DER)
    p7_bytes: bytes = file_get_data(file_data=p7s_data)
    if pem.detect(p7_bytes):
        _, _, p7_bytes = pem.unarmor(pem_bytes=p7_bytes)

    # parse the CMS structure
    err_msg: str | None = None
    signed_data: cms.SignedData | None = None
    try:
        content_info: cms.ContentInfo = cms.ContentInfo.load(encoded_data=p7_bytes)
        if content_info["content_type"].native == "signed_data":
            signed_data = content_info["content"]
        else:
            err_msg = "'p7_data' does not hold a valid PKCS#7 file"
    except Exception as e:
        err_msg = exc_format(exc=e,
                             exc_info=sys.exc_info())

    payload: bytes | None = None
    if not err_msg:
        # signatures in PKCS#7 are parallel, not chained, so they share the same payload
        embedded_content: bytes = signed_data["encap_content_info"]["content"].native

        # determine attached vs detached
        if embedded_content:
            # attached mode: use embedded content for digest
            payload = embedded_content
        elif doc_data:
            # detached mode: use external document
            payload = file_get_data(file_data=doc_data)
        else:
            err_msg = "For detached mode, a payload file must be provided"

    if not err_msg:
        # extract the signatures
        signer_infos: list[cms.SignerInfo] = signed_data["signer_infos"]

        # traverse the signatures
        for signer_info in signer_infos:
            # retrieve the signature properties
            signature_bytes: bytes = signer_info["signature"].native
            signature_algorithm: str = signer_info["signature_algorithm"]["algorithm"].native
            alg_name: str = signer_info["digest_algorithm"]["algorithm"].native
            hash_algorithm: HashAlgorithm = HashAlgorithm(alg_name)

            # extract the correct certificate
            signer_id: cms.SignerIdentifier = signer_info["sid"]
            signer_cert_bytes: bytes | None = None
            certs: cms.CertificateSet = signed_data["certificates"]
            for cert_choice in certs:
                # HAZARD: 'cert' is not a 'cryptography.x509.Certificate' object
                # (thus the need to dump it to DER bytes)
                cert: asn1crypto_x509.Certificate = cert_choice.chosen
                der_bytes: bytes = cert.dump()
                if signer_cert_bytes is None:
                    if signer_id.name == "issuer_and_serial_number":
                        # match issuer and serial number
                        if cert.issuer == signer_id.chosen["issuer"] and \
                           cert.serial_number == signer_id.chosen["serial_number"].native:
                            signer_cert_bytes = der_bytes
                    elif signer_id.name == "subject_key_identifier":
                        # extract SKI from certificate extensions
                        for ext in cert["tbs_certificate"]["extensions"]:
                            if ext["extn_id"].native == "subject_key_identifier" and \
                                    ext["extn_value"].native == signer_id.chosen.native:
                                signer_cert_bytes = der_bytes
                                break

            # extract the public key
            signer_cert: x509.Certificate = x509.load_der_x509_certificate(data=signer_cert_bytes)
            public_key: PublicKeyTypes = signer_cert.public_key()

            # extract the payload hash
            stored_hash: bytes | None = None
            signed_attrs: cms.CMSAttributes = signer_info["signed_attrs"]
            for signed_attr in signed_attrs:
                attr_type: str = signed_attr["type"].native
                if attr_type == "message_digest":
                    stored_hash = signed_attr["values"][0].native
                    break

            # validate the payload hash
            computed_hash: bytes = crypto_hash(msg=payload,
                                               alg=hash_algorithm)
            if not stored_hash:
                if logger:
                    logger.warning(msg="'p7s_data' has no stored payload digest")
            elif computed_hash != stored_hash:
                # hashes do not match
                err_msg = "Computed and stored digest values do not match"
                break

            # validate the signature
            chp_hash: ChpHash = _chp_hash(alg=hash_algorithm)
            if signed_attrs:
                # when 'signed_attrs' exists, the signature covers those attributes, not the raw data
                computed_hash = crypto_hash(msg=signed_attrs.dump(),
                                            alg=hash_algorithm)
            try:
                if isinstance(public_key, RSAPublicKey):
                    # determine the signature padding used
                    if "pss" in signature_algorithm:
                        signature_padding: padding.AsymmetricPadding = padding.PSS(
                            mgf=padding.MGF1(algorithm=chp_hash),
                            salt_length=padding.PSS.MAX_LENGTH
                        )
                    else:
                        signature_padding: padding.AsymmetricPadding = padding.PKCS1v15()

                    public_key.verify(signature=signature_bytes,
                                      data=computed_hash,
                                      padding=signature_padding,
                                      algorithm=Prehashed(chp_hash))
                else:
                    public_key.verify(signature=signature_bytes,
                                      data=computed_hash,
                                      algorithm=Prehashed(chp_hash))
                result = True
                if logger:
                    logger.debug(msg="Signature verification successful")
            except Exception as e:
                result = False
                if logger:
                    logger.warning(msg=f"Signature verification failed: {e}")
                break
    if err_msg:
        if logger:
            logger.error(msg=err_msg)
        if isinstance(errors, list):
            errors.append(err_msg)

    return result


def crypto_verify_pdf(pdf_data: Path | str | bytes,
                      cert_chain: Path | str | bytes = None,
                      errors: list[str] = None,
                      logger: Logger = None) -> bool | None:
    """
    Validate the embedded digital signatures of a PDF file in *PAdES* format.

    The nature of *pdf_data* depends on its data type:
      - type *bytes*: *pdf_data* holds the data (used as is)
      - type *str*: *pdf_data* holds the data (used as utf8-encoded)
      - type *Path*: *pdf_data* is a path to a file holding the data

    The folowing inconsistencies are verified, for each signature:
        - The digital signature is not valid
        - The certificate used has been revoked
        - The certificate used is not trusted
        - The signature block is not intact
        - A bad seed value was found

    The operation terminates on the first verification failure, regardless of the number of existing signatures.
    If *pdf_data* is not digitally signed, this is taken as an error.

    :param pdf_data: a PDF file path, or the PDF file bytes
    :param cert_chain: optional PEM/DER-encoded certificate chain
    :param errors: incidental error messages (may be non-empty)
    :param logger: optional logger
    :return: *True* if the signature is valid, *False* otherwise, or *None* if error
    """
    # initialize the return variable
    result: bool | None = None

    # obtain the PDF reader
    pdf_bytes: bytes = file_get_data(file_data=pdf_data)
    pdf_reader: PdfFileReader = PdfFileReader(stream=BytesIO(initial_bytes=pdf_bytes),
                                              strict=False)
    # obtain the validation context
    certs: Iterable[x509.Certificate] | None = None
    if cert_chain:
        certs_bytes: bytes = file_get_data(file_data=cert_chain)
        if certs_bytes:
            certs = load_certs_from_pemder_data(cert_data_bytes=certs_bytes)
    validation_context: ValidationContext = ValidationContext(trust_roots=certs)

    # obtain the list of digital signatures
    signatures: list[EmbeddedPdfSignature] = pdf_reader.embedded_signatures

    err_msg: str | None = None
    if signatures:
        # traverse the signatures
        result = True
        for signature in signatures:
            msg: str | None = None
            try:
                status: PdfSignatureStatus = validate_pdf_signature(embedded_sig=signature,
                                                                    signer_validation_context=validation_context)
                if status.revoked:
                    msg = "The certificate used has been revoked"
                elif not status.intact:
                    msg = "The signature block is not intact"
                elif not status.trusted:
                    msg = "The certificate used is not trusted"
                elif not status.seed_value_ok:
                    msg = "A bad seed value found"
                elif not status.valid:
                    msg = "The digital signature is not valid"

                if msg:
                    # an error has been flagged, report it
                    if logger:
                        logger.warning(msg=msg)
                    result = False
                    break

            except Exception as e:
                result = None
                err_msg = exc_format(exc=e,
                                     exc_info=sys.exc_info())
                break
    else:
        # signatures not retrieved, report the problem
        err_msg: str = "The file is not digitally signed"

    if err_msg:
        if logger:
            logger.error(msg=err_msg)
        if isinstance(errors, list):
            errors.append(err_msg)

    if result and logger:
        logger.debug(msg="Signature verification successful")

    return result



from __future__ import annotations  # allow forward references
import base64
import sys
from asn1crypto import cms, core, pem, tsp, x509 as asn1crypto_x509
from datetime import datetime
from dataclasses import dataclass
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7Options, PKCS7SignatureBuilder
from cryptography.hazmat.primitives.serialization.pkcs12 import load_key_and_certificates
from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
from logging import Logger
from pathlib import Path
from pypomes_core import file_get_data, exc_format
from typing import Any, Literal

from .crypto_common import (
    CRYPTO_DEFAULT_HASH_ALGORITHM,
    SignatureMode, HashAlgorithm, ChpHash, _chp_hash
)


class CryptoPkcs7:
    """
    Python code to extract crypto data from a PKCS#7 signature file.

    The crypto data is in *Cryptographic Message Syntax* (CMS), a standard for digitally signing, digesting,
    authenticating, andr encrypting arbitrary message content.

    These are the instance attributes:
        - p7s_bytes: bytes                 - the PKCS#7-compliant data in *DER* format
        - payload: bytes                   - the common payload (embedded or external)
        - signatures: list[SignatureInfo]  - data for list of signatures
    """
    # class-level logger
    logger: Logger | None = None

    @dataclass(frozen=True)
    class SignatureInfo:
        """
        These are the attributes holding the signature data.
        """
        payload_hash: bytes                 # the payload hash
        hash_algorithm: HashAlgorithm       # the algorithm used to calculate the payload hash
        signature: bytes                    # the digital signature
        signature_algorithm: str            # the algorithm used to generate the signature
        signature_timestamp: datetime       # the signature's timestamp
        public_key: PublicKeyTypes          # the public key (most likely, RSAPublicKey)
        signer_common_name: str             # the name of the certificate's signer
        signer_cert: x509.Certificate       # the reference certificate (latest one in the chain)
        cert_serial_number: int             # the certificate's serial nmumber
        cert_chain: list[bytes]             # the serialized X509 certificate chain (in DER format)

        # TSA (Time Stamping Authority) data
        tsa_timestamp: datetime             # the signature's timestamp
        tsa_policy: str                     # the TSA's policy
        tsa_serial_number: str              # the timestamping's serial number

    def __init__(self,
                 p7s_data: Path | str | bytes,
                 doc_data: Path | str | bytes = None,
                 errors: list[str] = None) -> None:
        """
        Instantiate the *CryptoPkcs7* class, and extract the relevant crypto data.

        The natures of *p7s_data* and *doc_data* depend on their respective data types:
          - type *bytes*: holds the data (used as is)
          - type *str*: holds the data (used as utf8-encoded)
          - type *Path*: is a path to a file holding the data

        The PKCS#7 data provided in *p7s_data* contains the A1 certificate and its corresponding
        public key, the certificate chain, the original payload (if *attached* mode), and the
        digital signature. The latter is always validated, and if a payload is specified in
        *doc_data* (*detached* mode), it is validated against its declared hash value.

        :param p7s_data: the PKCS#7 data in *DER* or *PEDM* format
        :param doc_data: the original document data (the payload, required in detached mode)
        :param errors: incidental errors
        """
        # declare/initialize the instance variables
        self.signatures: list[CryptoPkcs7.SignatureInfo] = []
        self.payload: bytes | None = None
        self.p7s_bytes: bytes

        # retrieve the PKCS#7 file data (if PEM, convert to DER)
        self.p7s_bytes = file_get_data(file_data=p7s_data)
        if pem.detect(self.p7s_bytes):
            _, _, self.p7s_bytes = pem.unarmor(pem_bytes=self.p7s_bytes)

        # define a local errors list
        curr_errors: list[str] = []

        # extract the base CMS structure
        signed_data: cms.SignedData | None = None
        try:
            content_info = cms.ContentInfo.load(encoded_data=self.p7s_bytes)
            if content_info["content_type"].native == "signed_data":
                signed_data = content_info["content"]
            else:
                msg = "'p7_data' does not hold a valid PKCS#7 file"
                if CryptoPkcs7.logger:
                    CryptoPkcs7.logger.error(msg=msg)
                curr_errors.append(msg)
        except Exception as e:
            msg: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
            if CryptoPkcs7.logger:
                CryptoPkcs7.logger.error(msg=msg)
            curr_errors.append(msg)

        if not curr_errors:
            # signatures in PKCS#7 are parallel, not chained, so they share the same payload
            encap_content: core.OctetString = signed_data["encap_content_info"]["content"]
            if encap_content:
                # attached mode
                self.payload = encap_content.native
            elif doc_data:
                # detached mode
                self.payload = file_get_data(file_data=doc_data)
            if not self.payload:
                msg = "For detached mode, a payload file must be provided"
                if CryptoPkcs7.logger:
                    CryptoPkcs7.logger.error(msg=msg)
                curr_errors.append(msg)

        if not curr_errors:
            # traverse the list of signatures
            signer_infos: list[cms.SignerInfo] = signed_data["signer_infos"]
            for signer_info in signer_infos:

                # extract the signature and its algorithms
                signature_bytes: bytes = signer_info["signature"].native
                signature_algorithm: str = signer_info["signature_algorithm"]["algorithm"].native
                alg_name: str = signer_info["digest_algorithm"]["algorithm"].native
                hash_algorithm: HashAlgorithm = HashAlgorithm(alg_name)

                stored_hash: bytes | None = None
                signature_timestamp: datetime | None = None
                signed_attrs: cms.CMSAttributes = signer_info["signed_attrs"] or []
                for signed_attr in signed_attrs:
                    attr_type: str = signed_attr["type"].native
                    match attr_type:
                        case "message_digest":
                            stored_hash = signed_attr["values"][0].native
                        case "signing_time":
                            signature_timestamp = signed_attr["values"][0].native

                # obtain/validate the hash
                from .crypto_pomes import crypto_hash
                computed_hash: bytes = crypto_hash(msg=self.payload,
                                                   alg=hash_algorithm)
                if not stored_hash:
                    if CryptoPkcs7.logger:
                        CryptoPkcs7.logger.warning(msg="'p7s_data' has no stored payload digest")
                    stored_hash = computed_hash
                elif computed_hash != stored_hash:
                    msg = "Computed and stored digest values do not match"
                    if CryptoPkcs7.logger:
                        CryptoPkcs7.logger.error(msg=msg)
                    curr_errors.append(msg)
                    break

                # select the correct certificate while building the certificate chain
                signer_id: cms.SignerIdentifier = signer_info["sid"]
                signer_cert_bytes: bytes | None = None
                cert_chain: list[bytes] = []
                certs: cms.CertificateSet = signed_data["certificates"]
                for cert_choice in certs:
                    # HAZARD: 'cert' is not a 'cryptography.x509.Certificate' object
                    cert: asn1crypto_x509.Certificate = cert_choice.chosen
                    der_bytes: bytes = cert.dump()
                    cert_chain.append(der_bytes)
                    if signer_cert_bytes is None:
                        if signer_id.name == "issuer_and_serial_number":
                            # match issuer and serial number
                            if cert.issuer == signer_id.chosen["issuer"] and \
                               cert.serial_number == signer_id.chosen["serial_number"].native:
                                signer_cert_bytes = der_bytes
                        elif signer_id.name == "subject_key_identifier":
                            # extract SKI from certificate extensions
                            for ext in cert["tbs_certificate"]["extensions"]:
                                if ext["extn_id"].native == "subject_key_identifier" and \
                                        ext["extn_value"].native == signer_id.chosen.native:
                                    signer_cert_bytes = der_bytes
                                    break

                # extract public key serial number
                signer_cert: x509.Certificate = x509.load_der_x509_certificate(data=cert_chain[0])
                public_key: PublicKeyTypes = signer_cert.public_key()
                cert_serial_number: int = signer_cert.serial_number

                # identify the signer
                subject: x509.name.Name = signer_cert.subject
                signer_common_name: str = subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value

                # TSA timestamp info (optional)
                tsa_timestamp: datetime | None = None
                tsa_policy: str | None = None
                tsa_serial_number: str | None = None

                unsigned_attrs: cms.CMSAttributes = signer_info["unsigned_attrs"]
                for unsigned_attr in unsigned_attrs:
                    attr_type: str = unsigned_attr["type"].native
                    if attr_type == "signature_time_stamp_token":
                        try:
                            # the timestamp token is a CMS SignedData structure -'dump()' gets the raw DER bytes
                            values: cms.SetOfContentInfo = unsigned_attr["values"]
                            timestamp_token: cms.ContentInfo = cms.ContentInfo.load(values[0].dump())

                            # extract the TSTInfo structure
                            tst_signed_data: cms.SignedData = timestamp_token["content"]
                            tst_info: tsp.TSTInfo = tst_signed_data["encap_content_info"]["content"].parsed

                            # extract TSA timestamp details
                            tsa_timestamp = tst_info["gen_time"].native
                            tsa_policy = tst_info["policy"].native
                            tsa_serial_number = hex(tst_info["serial_number"].native)

                        except Exception as e:
                            # unable to obtain TAS data: error parsing token
                            if CryptoPkcs7.logger:
                                msg: str = exc_format(exc=e,
                                                      exc_info=sys.exc_info())
                                CryptoPkcs7.logger.warning(msg=msg)
                        break

                # verify the signature
                if signed_attrs:
                    # when 'signed_attrs' exists, the signature covers those attributes, not the raw data
                    computed_hash = crypto_hash(msg=signed_attrs.dump(),
                                                alg=hash_algorithm)
                chp_hash: ChpHash = _chp_hash(alg=hash_algorithm)
                try:
                    if isinstance(public_key, RSAPublicKey):
                        # determine the signature padding used
                        if "pss" in signature_algorithm:
                            signature_padding: padding.AsymmetricPadding = padding.PSS(
                                mgf=padding.MGF1(algorithm=chp_hash),
                                salt_length=padding.PSS.MAX_LENGTH
                            )
                        else:
                            signature_padding: padding.AsymmetricPadding = padding.PKCS1v15()

                        public_key.verify(signature=signature_bytes,
                                          data=computed_hash,
                                          padding=signature_padding,
                                          algorithm=Prehashed(chp_hash))
                    else:
                        public_key.verify(signature=signature_bytes,
                                          data=computed_hash,
                                          algorithm=Prehashed(chp_hash))
                except Exception as e:
                    if CryptoPkcs7.logger:
                        msg: str = exc_format(exc=e,
                                              exc_info=sys.exc_info())
                        CryptoPkcs7.logger.warning(msg=msg + f" signed by {signer_common_name}")

                # build the signature's crypto data and save it
                sig_info: CryptoPkcs7.SignatureInfo = CryptoPkcs7.SignatureInfo(
                    payload_hash=stored_hash,
                    hash_algorithm=hash_algorithm,
                    signature=signature_bytes,
                    signature_algorithm=signature_algorithm,
                    signature_timestamp=signature_timestamp,
                    public_key=public_key,
                    signer_common_name=signer_common_name,
                    signer_cert=signer_cert,
                    cert_serial_number=cert_serial_number,
                    cert_chain=cert_chain,
                    tsa_timestamp=tsa_timestamp,
                    tsa_policy=tsa_policy,
                    tsa_serial_number=tsa_serial_number
                )
                self.signatures.append(sig_info)

        if curr_errors and isinstance(errors, list):
            errors.extend(curr_errors)

    def get_digest(self,
                   fmt: Literal["base64", "bytes"],
                   sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the digest associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the digest, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPkcs7.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.payload_hash \
            if fmt == "bytes" else base64.b64encode(s=sig_info.payload_hash).decode(encoding="utf-8")

    def get_signature(self,
                      fmt: Literal["base64", "bytes"],
                      sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the signature associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the signature, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPkcs7.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.signature \
            if fmt == "bytes" else base64.b64encode(s=sig_info.signature).decode(encoding="utf-8")

    def get_public_key(self,
                       fmt: Literal["base64", "der", "pem"],
                       sig_seq: int = 0) -> str | bytes:
        """
        Retrieve the public key associated with a reference signature, as specified in *sig_seq* and *fmt*.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        These are the supported formats:
            - *der*: the raw binary representation of the key
            - *pem*: the Base64-encoded key with headers and line breaks
            - *base64*: the Base64-encoded DER bytes

        :param fmt: the format to use
        :param sig_seq: the relative ordinal position of the reference signature
        :return: the public key, as per *fmt* (*str* or *bytes*)
        """
        # declare the return variable
        result: str | bytes

        sig_info: CryptoPkcs7.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        if fmt == "pem":
            result = sig_info.public_key.public_bytes(encoding=Encoding.PEM,
                                                      format=PublicFormat.SubjectPublicKeyInfo)
            result = result.decode(encoding="utf-8")
        else:
            result = sig_info.public_key.public_bytes(encoding=Encoding.DER,
                                                      format=PublicFormat.SubjectPublicKeyInfo)
            if fmt == "base64":
                result = base64.b64encode(s=result).decode(encoding="utf-8")

        return result

    def get_cert_chain(self,
                       sig_seq: int = 0) -> list[bytes]:
        """
        Retrieve the certificate chain associated with a reference signature, as specified in *sig_seq**.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the signature, as per *fmt* (Base64-encoded or raw bytes)
        """
        sig_info: CryptoPkcs7.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        return sig_info.cert_chain

    def get_metadata(self,
                     sig_seq: int = 0) -> dict[str, Any]:
        """
        Retrieve the certificate chain metadata associated with a reference signature, as specified in *sig_seq*.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the certificate chain metadata associated with the reference signature
        """
        # declare the return variable
        result: dict[str, Any]

        sig_info: CryptoPkcs7.SignatureInfo = self.__get_sig_info(sig_seq=sig_seq)
        cert: x509.Certificate = sig_info.signer_cert

        result: dict[str, Any] = {
            "signer-common-name": sig_info.signer_common_name,
            "hash-algorithm": sig_info.hash_algorithm,
            "signature-algorithm": sig_info.signature_algorithm,
            "signature-timestamp": sig_info.signature_timestamp,
            "cert-serial-number": sig_info.cert_serial_number,
            "cert-not-before": cert.not_valid_before,
            "cert-not-after": cert.not_valid_after,
            "cert-subject": cert.subject.rfc4514_string(),
            "cert-issuer": cert.issuer.rfc4514_string(),
            "cert-chain-length": len(sig_info.cert_chain)
        }
        # add the TSA details
        if sig_info.tsa_serial_number:
            result.update({
                "tsa-timestamp": sig_info.tsa_timestamp,
                "tsa-policy": sig_info.tsa_policy,
                "tsa-serial-number": sig_info.tsa_serial_number
            })

        return result

    def __get_sig_info(self,
                       sig_seq: int) -> CryptoPkcs7.SignatureInfo:
        """
        Retrieve the signature metadata of a reference signature, as specified in *sig_seq*.

        The natural ordering of the signatures in a *PKCS#7* compliant *.p7s* file is the chronological
        *latest-first* order. The value of *sig_seq* is subtracted from the ordinal position of the last
        signature in the signatures list, to yield the ordinal position of the reference signature.
        It defaults to *0*, indicating the latest signature. If the operation yields a number out of
        the range of available signatures, the latest signature is selected.

        :param sig_seq: the relative ordinal position of the reference signature
        :return: the reference signature's metadata

        """
        sig_ordinal: int = max(-1, len(self.signatures) - sig_seq - 1)
        return self.signatures[sig_ordinal]

    @staticmethod
    def create(doc_data: Path | str | bytes,
               pfx_data: Path | str | bytes,
               pfx_pwd: str | bytes = None,
               p7s_out: Path | str = None,
               hash_alg: HashAlgorithm = CRYPTO_DEFAULT_HASH_ALGORITHM,
               sig_mode: SignatureMode = SignatureMode.DETACHED,
               errors: list[str] = None) -> CryptoPkcs7:
        """
        Instantiate a *CryptoPkcs7* object by signing a document with a type A1 certificate.

        The natures of *doc_data* and *pfx_data* depend on their respective data types:
          - type *bytes*: holds the data (used as is)
          - type *str*: holds the data (used as utf8-encoded)
          - type *Path*: is a path to a file holding the data

        The signature is created as a PKCS#7/CMS compliant structure with full certificate chain.
        The parameter *sig_mode* determines whether the payload is to be embedded (*attached*),
        or left aside (*detached*).

        :param doc_data: the document to sign
        :param pfx_data: the PKCS#12 (*.pfx*) data, containing A1 certificate and private key
        :param pfx_pwd: password for the *.pfx* data (if not provided, *pfx_data* is assumed to be unencrypted)
        :param p7s_out: path to the output PKCS#7 file (optional, no output if not provided)
        :param hash_alg: the algorithm for hashing
        :param sig_mode: whether to handle the payload as "attached" (defaults to "detached")
        :param errors: incidental errors (may be non-empty)
        :return: the instance of *CryptoPkcs7*, or *None* if error
        """
        # initialize the return variable
        result: CryptoPkcs7 | None = None

        # definal a local errors list
        curr_errors: list[str] = []

        # retrieve the document and certificate raw bytes
        doc_bytes: bytes = file_get_data(file_data=doc_data)
        pfx_bytes: bytes = file_get_data(file_data=pfx_data)

        # load A1 certificate and private key from the raw certificate data
        pwd_bytes = pfx_pwd.encode() if isinstance(pfx_pwd, str) else pfx_pwd
        cert_data: tuple = load_key_and_certificates(data=pfx_bytes,
                                                     password=pwd_bytes)
        private_key: PrivateKeyTypes = cert_data[0]
        cert_main: x509.Certificate = cert_data[1]
        sig_hasher: ChpHash = _chp_hash(alg=hash_alg,
                                        errors=curr_errors)

        if cert_main and private_key and sig_hasher:
            additional_certs: list[x509.Certificate] = cert_data[2] or []

            # prepare the PKCS#7 builder
            builder: PKCS7SignatureBuilder = PKCS7SignatureBuilder(data=doc_bytes)
            builder = builder.add_signer(certificate=cert_main,
                                         private_key=private_key,
                                         hash_algorithm=sig_hasher)
            # add full certificate chain to the return data
            for cert in additional_certs:
                builder = builder.add_certificate(cert)

            # define PKCS#7 options
            options: list[PKCS7Options] = [PKCS7Options.Binary]
            if sig_mode == SignatureMode.DETACHED:
                options.append(PKCS7Options.DetachedSignature)

            # build the PKCS#7 data in DER format
            pkcs7_data: bytes = builder.sign(encoding=Encoding.DER,
                                             options=options)
            # instantiate the object
            result = CryptoPkcs7(p7s_data=pkcs7_data,
                                 doc_data=doc_bytes if sig_mode == SignatureMode.DETACHED else None,
                                 errors=curr_errors)

            # output the PKCS#7 file
            if not curr_errors and p7s_out:
                # make sure 'p7s_out' is a 'Path'
                p7s_out = Path(p7s_out)
                # write the PKCS#7 data to a file
                with p7s_out.open("wb") as out_f:
                    out_f.write(pkcs7_data)

        elif not curr_errors:
            if not cert_main:
                msg: str = "Failed to load the digital certificate"
            else:
                msg: str = "Failed to load the private key"
            if CryptoPkcs7.logger:
                CryptoPkcs7.logger.error(msg=msg)
            curr_errors.append(msg)

        if curr_errors and isinstance(errors, list):
            errors.extend(curr_errors)

        return result

    @staticmethod
    def set_logger(logger: Logger) -> None:
        """
        Configure the logger to be used in this module's operations.

        :param logger: the operations logger
        """
        CryptoPkcs7.logger = logger



import base64
import jwt
import requests
import sys
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from jwt import PyJWK
from jwt.algorithms import RSAPublicKey
from logging import Logger
from pypomes_core import exc_format
from typing import Any, Literal


def jwt_convert(jwk: dict[str, str],
                fmt: Literal["DER", "PEM"]) -> bytes | str | None:
    """
    Convert the *JWK* (JSON Web Key) given in *jwk* to the format *fmt*.

    The supported target formats are:
        - *DER*: Distinguished Encoding Rules (bytes)
        - *PEM*: Privacy-Enhanced Mail (str)

    A typical JWK has the following format (for simplicity, 'n' and 'x5c' are truncated):
    {
        "kid": "5vrDgXBKmE-7EorvkE5Mu9VlId_ISgl2v7Af23nDvDU",
        "kty": "RSA",
        "alg": "RS256",
        "use": "sig",
        "n": "2lfctxbCT2UWsBMC1rlpmWRcmDRS-EN0opirHzUsgHf7VJCyN0beKchy3biKQeXhgoirWR-f9uufn6Wl...",
        "e": "AQAB",
        "x5c": [
            "MIIClTCCAX0CBgFxx6eIWjANBgkqhkiG9w0BAQsFADAOMQwwCgYDVQQDDANwamUwHhcNMjAwNDI5MjAzNDM3..."
        ],
        "x5t": "Bq0yqAX_D4aFA0eX9HSBVZxVW3A",
        "x5t#S256": "OGHvtCjTBasA9uHivO1_cNJXKExc0w1-1yhTPEK2CPM"
    }

    :param jwk: the JSON web key to be converted
    :param fmt: the target format
    :return: *jwt* in *DER* (bytes) or *PEM* (str) format, or *None* if error
    """
    # retrieve and decode the base64url values
    n_b64: str = jwk["n"]
    e_b64: str = jwk["e"]
    n_int: int = int.from_bytes(base64.urlsafe_b64decode(n_b64 + "=="), "big")
    e_int: int = int.from_bytes(base64.urlsafe_b64decode(e_b64 + "=="), "big")

    # construct the RSA public key
    public_numbers = rsa.RSAPublicNumbers(e=e_int,
                                          n=n_int)
    public_key = public_numbers.public_key(backend=default_backend())

    result: bytes | str | None = None
    if fmt == "DER":
        # export to DER format
        result: bytes = public_key.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
    elif fmt == "PEM":
        # export to PEM format
        pem_bytes: bytes = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        result: str = pem_bytes.decode("utf-8")

    return result


def jwt_get_header(token: str,
                   errors: list[str] = None,
                   logger: Logger = None) -> dict[str, Any] | None:
    """
    Retrieve the data in the header a JWT *token*.

    Any well-constructed JWT token may be provided in *token*.
    Note that neither the token's signature nor its expiration is verified.

    Structure of the returned data:
        {
            "alg": "RS256",
            "typ": "JWT",
            "kid": "A1234"
        }

    :param token: the reference token
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the data in the token's header, or *None* if error
    """
    # initialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg="Retrieve header for token")

    try:
        result = jwt.get_unverified_header(jwt=token)
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error retrieving the token's header: {exc_err}")
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def jwt_get_payload(token: str,
                    errors: list[str] = None,
                    logger: Logger = None) -> dict[str, dict[str, Any]] | None:
    """
    Retrieve the data in the payload of a JWT *token*.

    Any well-constructed JWT token may be provided in *token*.
    Note that neither the token's signature nor its expiration is verified.

    Structure of the returned data:
        {
            "birthdate": "1980-01-01",
            "email": "jdoe@mail.com",
            "exp": 1516640454,
            "iat": 1516239022,
            "iss": "my_jwt_provider.com",
            "jti": "Uhsdfgr67FGH567qwSDF33er89retert",
            "gender": "M",
            "name": "John Doe",
            "nbt": 1516249022
            "sub": "11111111111",
            "roles": [
                "administrator",
                "operator"
            ]
      }

    :param token: the reference token
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the data in the token's payload, or *None* if error
    """
    # initialize the return variable
    result: dict[str, dict[str, Any]] | None = None

    if logger:
        logger.debug(msg="Retrieve payload for token")

    try:
        result = jwt.decode(jwt=token,
                            options={"verify_signature": False})
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error retrieving the token's payload: {exc_err}")
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def jwt_get_claim(token: str,
                  key: str,
                  errors: list[str] = None,
                  logger: Logger = None) -> str | None:
    """
    Retrieve the claim associated with *key* in *token*'s header or payload.

    :param token: the reference token
    :param key: the name of the claim whose values are to be returned
    :param errors: incidental errors
    :param logger: optiona logger
    :return: the requested claim, or *None* if error or not found
    """
    # initialize the return variable:
    result: str | None = None

    header: dict[str, Any] = jwt_get_header(token=token,
                                            errors=errors,
                                            logger=logger)
    if not errors:
        result = header.get(key)
        if not result:
            payload: dict[str, Any] = jwt_get_payload(token=token,
                                                      errors=errors,
                                                      logger=logger)
            if not errors:
                result = payload.get(key)

    return result


def jwt_get_claims(token: str,
                   keys: tuple[str, ...],
                   errors: list[str] = None,
                   logger: Logger = None) -> tuple | None:
    """
    Retrieve the claims associated with *keys* in *token*'s header or payload.

    The claims are returned in the same order as requested in *keys*.
    For a claim not found, *None* is returned in its position.

    :param token: the reference token
    :param keys: the names of the claims whose values are to be returned
    :param errors: incidental errors
    :param logger: optiona logger
    :return: a tuple containing the respective values of *keys* in *token*'s payload, or *None* if error
    """
    # initialize the return variable:
    result: tuple | None = None

    header: dict[str, Any] = jwt_get_header(token=token,
                                            errors=errors,
                                            logger=logger)
    if not errors:
        payload: dict[str, Any] = jwt_get_payload(token=token,
                                                  errors=errors,
                                                  logger=logger)
        if not errors:
            result = tuple([header.get(key) or payload.get(key) for key in keys])

    return result


def jwt_get_public_key(issuer: str = None,
                       token: str = None,
                       fmt: Literal["PEM", "DER", "JWK"] = "PEM",
                       errors: list[str] = None,
                       logger: Logger = None) -> dict[str, str] | bytes | str | None:
    """
    Obtain the public key used to sign the token.

    The token's issuer may be provided in one of these two parameters:
        - *issuer*: the content of the token claim *iss*
        - *token*: the token itself, from which the *issuer* is extracted

    This is accomplished by requesting the token issuer for its *JWKS* (JSON Web Key Set),
    containing the public keys used for various purposes, as indicated in the attribute *use*:
        - *enc*: the key is intended for encryption
        - *sig*: the key is intended for digital signature
        - *wrap*: the key is intended for key wrapping

    A typical JWKS set has the following format (for simplicity, 'n' and 'x5c' are truncated):
        {
            "keys": [
                {
                    "kid": "X2QEcSQ4Tg2M2EK6s2nhRHZH_GwD_zxZtiWVwP4S0tg",
                    "kty": "RSA",
                    "alg": "RSA256",
                    "use": "sig",
                    "n": "tQmDmyM3tMFt5FMVMbqbQYpaDPf6A5l4e_kTVDBiHrK_bRlGfkk8hYm5SNzNzCZ...",
                    "e": "AQAB",
                    "x5c": [
                        "MIIClzCCAX8CBgGZY0bqrTANBgkqhkiG9w0BAQsFADAPMQ0wCwYDVQQDDARpanVk..."
                    ],
                    "x5t": "MHfVp4kBjEZuYOtiaaGsfLCL15Q",
                    "x5t#S256": "QADezSLgD8emuonBz8hn8ghTnxo7AHX4NVNkr4luEhk"
                },
                ...
            ]
        }

    The signature key is returned in its original *JWK* (JSON Web Key) format, or converted to
    either *DER* (Distinguished Encoding Rules) or *PEM* (Privacy-Enhanced Mail) format, as per *ftm*.

    :param issuer: the token's issuer, as presented in the token's *iss* claim (required, if *token* is not provided)
    :param token: the reference token (required, if *issuer* is not provided)
    :param fmt: the returning key's format (defaults to *PEM*)
    :param errors: incidental error messages
    :param logger: optional logger
    :return: the public key in *JWT*, *DER*, or *PEM* format, or *None* if error
    """
    # initialize the return variable
    result: dict[str, str] | bytes | str | None = None

    if not issuer and token:
        issuer: str = jwt_get_claim(token=token,
                                    key="iss",
                                    errors=errors,
                                    logger=logger)
    if issuer:
        # obtain the JWKS (JSON Web Key Set) from the token issuer
        url: str = f"{issuer}/protocol/openid-connect/certs"
        if logger:
            logger.debug(msg=f"GET {url}")
        try:
            response: requests.Response = requests.get(url=url)
            if response.status_code == 200:
                # request succeeded
                if logger:
                    logger.debug(msg="GET success")
                # select the appropriate JWK
                reply: dict[str, list[dict[str, str]]] = response.json()
                jwk: dict[str, str] | None = None
                for key in reply["keys"]:
                    if key.get("use") == "sig":
                        jwk = key
                        break
                if jwk:
                    # convert from 'JWK' to 'PEM' and save it for further use
                    if fmt in ["DER", "PEM"]:
                        # noinspection PyTypeChecker
                        result = jwt_convert(jwk=jwk,
                                             fmt=fmt)
                    else:
                        result = jwk
                    if result and logger:
                        logger.debug(f"Public key obtained for issuer '{issuer}'")
                else:
                    msg: str = (f"Signature public key missing from the JWKS "
                                f"returned by the token issuer '{issuer}'")
                    if logger:
                        logger.error(msg=msg)
                    if isinstance(errors, list):
                        errors.append(msg)
            elif logger:
                msg: str = f"GET failure, status {response.status_code}, reason {response.reason}"
                if hasattr(response, "content") and response.content:
                    msg += f", content {response.content}"
                logger.error(msg=msg)
                if isinstance(errors, list):
                    errors.append(msg)
        except Exception as e:
            # the operation raised an exception
            msg = exc_format(exc=e,
                             exc_info=sys.exc_info())
            if logger:
                logger.error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)

    elif not errors:
        msg = "Token/issuer not provided"
        if logger:
            logger.error(msg=msg)
        if isinstance(errors, list):
            errors.append(msg)

    return result


def jwt_validate(token: str,
                 issuer: str = None,
                 recipient_id: str = None,
                 recipient_attr: str = None,
                 public_key: str | bytes | PyJWK | RSAPublicKey = None,
                 errors: list[str] = None,
                 logger: Logger = None) -> dict[str, dict[str, Any]] | None:
    """
    Verify whether *token* is a valid JWT token, and return its claims (sections *header* and *payload*).

    The supported public key types are:
        - *DER*: Distinguished Encoding Rules (bytes)
        - *PEM*: Privacy-Enhanced Mail (str)
        - *PyJWK*: a formar from the *PyJWT* package
        - *RSAPublicKey*: a format from the *PyJWT* package

    If an asymmetric algorithm was used to sign the token and *public_key* is provided, then
    the token's signature is validated, by using the data in its *signature* section.

    The parameters *recipient_id* and *recipient_attr* refer the token's expected subject, respectively,
    the subject's identification and the attribute in the token's payload data identifying its subject.
    If both are provided, *recipient_id* is validated.

    On failure, *errors* will contain the reason(s) for rejecting *token*.
    On success, return the token's claims (*header* and *payload*).

    :param token: the token to be validated
    :param public_key: optional public key used to sign the token, in *PEM* format
    :param issuer: optional value to compare with the token's *iss* (issuer) attribute in its *payload*
    :param recipient_id: identification of the expected token subject
    :param recipient_attr: attribute in the token's payload holding the expected subject's identification
    :param errors: incidental error messages
    :param logger: optional logger
    :return: The token's claims (*header* and *payload*), or *None* if error
    """
    # initialize the return variable
    result: dict[str, dict[str, Any]] | None = None

    if logger:
        logger.debug(msg="Validate JWT token")

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # extract needed data from token header
    token_header: dict[str, Any] | None = None
    try:
        token_header: dict[str, Any] = jwt.get_unverified_header(jwt=token)
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if logger:
            logger.error(msg=f"Error retrieving the token's header: {exc_err}")
        errors.append(exc_err)

    # validate the token
    if not errors:
        token_alg: str = token_header.get("alg")
        require: list[str] = ["exp", "iat"]
        if issuer:
            require.append("iss")
        options: dict[str, Any] = {
            "require": require,
            "verify_aud": False,
            "verify_exp": True,
            "verify_iat": True,
            "verify_iss": issuer is not None,
            "verify_nbf": False,
            "verify_signature": token_alg in ["RS256", "RS512"] and public_key is not None
        }
        try:
            # raises:
            #   InvalidTokenError: token is invalid
            #   InvalidKeyError: authentication key is not in the proper format
            #   ExpiredSignatureError: token and refresh period have expired
            #   InvalidSignatureError: signature does not match the one provided as part of the token
            #   ImmatureSignatureError: 'nbf' or 'iat' claim represents a timestamp in the future
            #   InvalidAlgorithmError: the specified algorithm is not recognized
            #   InvalidIssuedAtError: 'iat' claim is non-numeric
            #   MissingRequiredClaimError: a required claim is not contained in the claimset
            payload: dict[str, Any] = jwt.decode(jwt=token,
                                                 key=public_key,
                                                 algorithms=[token_alg],
                                                 options=options,
                                                 issuer=issuer)
            if recipient_id and recipient_attr and \
                    payload.get(recipient_attr) and recipient_id != payload.get(recipient_attr):
                msg: str = f"Token was issued to '{payload.get(recipient_attr)}', not to '{recipient_id}'"
                if logger:
                    logger.error(msg=msg)
                errors.append(msg)
            else:
                result = {
                    "header": token_header,
                    "payload": payload
                }
        except Exception as e:
            exc_err: str = exc_format(exc=e,
                                      exc_info=sys.exc_info())
            if logger:
                logger.error(msg=f"Error decoding the token: {exc_err}")
            errors.append(exc_err)

    if logger:
        if errors:
            logger.debug(msg=f"Token is invalid: {token}")
        else:
            logger.debug(msg="Token is valid")

    return result



from cryptography.hazmat.primitives import hashes
from enum import StrEnum, auto
from logging import Logger
from pypomes_core import env_get_enum, APP_PREFIX
from typing import Final


class SignatureMode(StrEnum):
    """
    Location of signatures with respect to the signed files.
    """
    ATTACHED = auto()
    DETACHED = auto()


class SignatureType(StrEnum):
    """
    Types of cryptographic signatures in documents.
    """
    CADES = "CAdES"
    PADES = "PAdES"
    XADES = "XAdES"


class HashAlgorithm(StrEnum):
    """
    Supported hash algorithms.
    """
    MD5 = auto()
    BLAKE2B = auto()
    BLAKE2S = auto()
    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SHA3_224 = auto()
    SHA3_256 = auto()
    SHA3_384 = auto()
    SHA3_512 = auto()
    SHAKE_128 = auto()
    SHAKE_256 = auto()


ChpHash = hashes.SHA224 | hashes.SHA256 | hashes.SHA384 | hashes.SHA512

CRYPTO_DEFAULT_HASH_ALGORITHM: Final[HashAlgorithm] = \
    env_get_enum(key=f"{APP_PREFIX}_CRYPTO_DEFAULT_HASH_ALGORITHM",
                 enum_class=HashAlgorithm,
                 def_value=HashAlgorithm.SHA256)


def _chp_hash(alg: HashAlgorithm | str,
              errors: list[str] = None,
              logger: Logger = None) -> ChpHash:
    """
    Construct the *cryptography* package's hash object corresponding top *hash_alg*.

    The hash object is an instance of *cryptography.hazmat.primitives.hashes.<hash>*

    :param alg: the hash algorithm
    :param errors: incidental errors
    :return: the *Crypto* package's hash object, or *None* if error
    """
    result: ChpHash | None = None
    match alg:
        case HashAlgorithm.SHA224:
            result = hashes.SHA224()
        case HashAlgorithm.SHA256:
            result = hashes.SHA256()
        case HashAlgorithm.SHA384:
            result = hashes.SHA384()
        case HashAlgorithm.SHA512:
            result = hashes.SHA512()
        case _:
            msg = f"Hash algorithm not supported: '{alg}'"
            if logger:
                logger.error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)

    return result
